{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFClientDataSet.pas - TghClientDataSet class unit.                     }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFClientDataSet;  { Client Data Set }

Interface

  Uses
    GHFCustomSavePoint, GHFCustomPurposeSavePoints, DBClient, Contnrs,
    GHFVCL, DB, Provider, DSIntf, GHFRTL, Classes, GHFDataSource,
    GHFObjList;

  Type
    { Client Data Set Save Point class }
    TghClientDataSetSavePoint = Class (TghCustomSavePoint)
      Protected
        { Fields }
        State :Integer;

        { Overridden methods }
        Procedure InternalConfirm; Override;
        Procedure InternalRestore; Override;
        Procedure InternalSave; Override;
      Public
        { Overridden methods }
        Procedure Invalidate (
          DetailAction :TghSavePointAction = ghsaNone); Override;
    End;

    { Client Data Set Save Points class }
    TghClientDataSetSavePoints = Class (TghCustomPurposeSavePoints)
      Protected
        { Overridden methods }
        Function NewPointClass :TghCustomSavePointClass; Override;
    End;

    { TghClientDataSetSavePoints class type }
    TghClientDataSetSavePointsClass = Class Of
      TghClientDataSetSavePoints;

    { Client Data Set class }
    TghClientDataSet = Class (TClientDataSet)
      Protected
        { Fields }
        ApplyingInternalPost :Boolean;
        BlockReadSizeStack :TStack;
        ChangeNotificationLockCount :Integer;
        ChangeNotificationReg :Boolean;
        CheckingDetailRecs :Boolean;
        FAfterFieldChange :TghFieldAfterChangeEvent;
        FAllParamsWhenDetail :Boolean;
        FAutoApplyDetails :Boolean;
        FAutoCancelDetails :Boolean;
        FAutoDeleteDetails :Boolean;
        FAutoEdit :Boolean;
        FBeforeFieldChange :TghFieldBeforeChangeEvent;
        FBeforeFocusField :TghBeforeContinueFieldEvent;
        FCanceling :Boolean;
        FDeleting :Boolean;
        FDetailList :TObjectList;
        FFieldParamNames :String;
        FID :String;
        FIdentify :Boolean;
        FImmediateApplyUpdates :Boolean;
        FIsAuxClone :Boolean;
        FMasterEventOnOpen :Boolean;
        FPersistentState :TDataSetState;
        FPosting :Boolean;
        FPrevPersistentState :TDataSetState;
        FProvider :TCustomProvider;
        FSavePoints :TghClientDataSetSavePoints;
        FStoreActive :Boolean;
        FUpdateErrors :Integer;
        FWaitingForApplyingConfirmation :Boolean;
        InChangeNotification :Boolean;
        InternalDeleting :Boolean;
        QuietDeletingCount :Integer;
        ResyncLockCount :Integer;
        TempFieldData :PPointer;
        UpdatingCursor :Boolean;
        UpdatingState :Boolean;

        { Static methods }
        Function ActiveRecBookmark :PDSBookmrk;
        Function ActiveRecBuffer :PChar;
        Function BaseProp (Const Prop :DSProp) :Integer;
        Function ChangeCurrentRec (Const Cursor :IDSCursor;
          Const AllowCalcs :Boolean) :Boolean;
        Function ChangeIndexAttrs :Integer;
        Procedure ChangeNotification; StdCall;
        Function ChangeNotificationLocked :Boolean;
        Procedure CheckDeleteDetails;
        Procedure CheckIdentify (Var Data :OLEVariant);
        Procedure CheckRemoveDetails;
        Procedure CheckRevertDetails;
        Function CheckStatus (Const Status, Value :DBResult) :Boolean;
        Procedure CheckWaitingForApplyingConfirmation;
        Function Contains (DataSet :TghClientDataSet) :Boolean;
        Function CreateCursor (Const ABookmark :TBookmark = Nil)
          :IDSCursor;
        Function CurrentBookmark :DSBookmrk;
        Function CursorRecAttrs (Const Cursor :IDSCursor) :DSAttr;
        Function CursorRecCount (Const Cursor :IDSCursor) :Integer;
        Function CursorRecNumber (Const Cursor :IDSCursor) :Cardinal;
        Procedure DeleteRecsDetail (Const AMaster :TDataSet;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant);
        Procedure DeleteRecsFilter (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant);
        Procedure DeleteRecsNested (Const ParentRec :Cardinal;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant);
        Procedure DeleteRecsVisible (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant);
        Function DetailList (Const Update :Boolean = False) :TObjectList;
        Function FieldDesc (Const Field :String) :DSFLDDesc;
        Function FindSavePoints :TghClientDataSetSavePoints;
        Function GetAutoApplyDetails :Boolean;
        Function GetCanceling :Boolean;
        Function GetDeleting :Boolean;
        Function GetDetail (Const Index :Integer) :TDataSet;
        Function GetDetailed :Boolean;
        Function GetFieldData (Const Cursor :IDSCursor;
          Const RecBuffer :Pointer; Const Field :Integer;
          Const Data :Pointer = Nil) :Boolean; Overload;
        Function GetFieldData (Const RecBuffer :Pointer;
          Const Field :Integer; Const Data :Pointer = Nil) :Boolean;
          Overload;
        Function GetFieldDescs :TFieldDescList;
        Procedure GetIndexFieldList (Const List :TList);
        Function GetIsChangeLog :Boolean;
        Function GetLimited :Boolean;
        Function GetMaster :TDataSet;
        Function GetMasterActive :Boolean;
        Procedure GetMasterFieldList (Const List :TList);
        Function GetNested (Const Index :Integer) :TghClientDataSet;
        Function GetNestedCount :Integer;
        Function GetParent :TghClientDataSet;
        Function GetParentRecNumber :Cardinal;
        Function GetPosting :Boolean;
        Function GetPrimary :TghClientDataSet;
        Function GetRanged :Boolean;
        Function GetRecNumber :Cardinal;
        Function GetRootMaster :TghClientDataSet;
        Function GetRootParent :TghClientDataSet;
        Function GetSavePoints :TghClientDataSetSavePoints;
        Function GetSortFieldNames :String;
        Function GetStatusFilterAttrs :Integer;
        Function HasSavePoint (Const SavePoint :TghCustomSavePoint)
          :Boolean;
        Function InternalGetRecNumber :Cardinal;
        Function InternalID :WideString;
        Function InvalidateNestedSavePoints (
          Const StartGlobalIndex :Integer = 0) :Integer;
        Procedure LockChangeNotification;
        Procedure LockResync;
        Function NativeValue (Const Field :TField; Const Buffer :Pointer)
          :Variant;
        Function RecBookmark (Const RecBuffer :Pointer) :PDSBookmrk;
        Function RecInfo (Const RecBuffer :Pointer) :PRecInfo;
        Procedure RefreshChangeLogs (Const IncludeSelf :Boolean = True);
        Procedure RegChangeNotification;
        Function ResyncLocked :Boolean;
        Procedure RevertRecsDetail (Const AMaster :TDataSet;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant);
        Procedure RevertRecsFilter (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant);
        Procedure RevertRecsNested (Const ParentRec :Cardinal;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant);
        Procedure RevertRecsVisible (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant);
        Function RowCountDetail (Const AMaster :TDataSet;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant)
          :Integer;
        Function RowCountFilter (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant) :Integer;
        Function RowCountNested (Const ParentRec :Cardinal;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant)
          :Integer;
        Function RowCountVisible (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant) :Integer;
        Function SameBase (Const DataSetIndex :Integer) :Boolean;
        Procedure SetBaseOrder (Const Cursor :IDSCursor);
        Procedure SetFieldData (Const Cursor :IDSCursor;
          Const RecBuffer :Pointer; Const Field :Integer;
          Const Data :Pointer = Nil); Overload;
        Procedure SetFieldData (Const RecBuffer :Pointer;
          Const Field :Integer; Const Data :Pointer = Nil); Overload;
        Procedure SetImmediateApplyUpdates (Const Value :Boolean);
        Procedure SetInternalBlankFlag (Const Rec, Field :Integer;
          Const Value :Byte);
        Procedure SetLogChanges (Const Value :Boolean);
        Procedure SetProviderName (Const Value :String);
        Procedure SetProviderProp (Const Value :TCustomProvider);
        Function SetUpdatingCursor (Const Value :Boolean = True) :Boolean;
        Procedure SetWaitingForApplyingConfirmation (Const Value :Boolean);
        Procedure UnLockChangeNotification;
        Procedure UnLockResync;
        Procedure UpdateChangeIndex (Attrs :Integer = -1);

        { Virtual methods }
        Function CanCheckFieldChange (Field :TField) :Boolean; Virtual;
        Procedure DoAfterFieldChange (Field :TField;
          Const PrevValue :Variant); Virtual;
        Procedure DoBeforeFieldChange (Field :TField;
          Const NewValue :Variant); Virtual;
        Function DoRecsDetail (AMaster :TDataSet; Method :TghDoEnumMethod;
          FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
          Virtual;
        Function DoRecsNested (ParentRec :Cardinal; Method :TghDoEnumMethod;
          FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
          Virtual;
        Function GetChangeLog :TghClientDataSet; Virtual;
        Procedure InternalChangeNotification; Virtual;
        Procedure InternalMergeChangeLog; Virtual;
        Function LinkDataSourceClass :TghDataSourceClass; Virtual;
        Function MustNotifyFieldChange (Field :TField) :Boolean; Virtual;
        Function SavePointsClass :TghClientDataSetSavePointsClass;
          Virtual;
        Procedure SetStatusFilter (Const Value :TUpdateStatusSet); Virtual;

        { Overridden methods }
        Procedure AddDataPacket (Const Data :OLEVariant; HitEOF :Boolean);
          Override;
        Procedure CheckDetailRecords; Override;
        Procedure CloseCursor; Override;
        Procedure DataEvent (Event :TDataEvent; Info :Longint); Override;
        Procedure DoAfterDelete; Override;
        Procedure DoAfterOpen; Override;
        Procedure DoAfterPost; Override;
        Function DoApplyUpdates (Delta :OLEVariant; MaxErrors :Integer;
          Out ErrorCount :Integer) :OLEVariant; Override;
        Procedure DoBeforeApplyUpdates (Var OwnerData :OLEVariant);
          Override;
        Procedure DoBeforeDelete; Override;
        Procedure DoBeforeEdit; Override;
        Procedure DoBeforeExecute (Var OwnerData :OLEVariant); Override;
        Procedure DoBeforeGetParams (Var OwnerData :OLEVariant); Override;
        Procedure DoBeforeGetRecords (Var OwnerData :OLEVariant); Override;
        Procedure DoBeforeInsert; Override;
        Procedure DoBeforeRowRequest (Var OwnerData :OLEVariant); Override;
        Function DoGetRecords (Count :Integer; Out RecsOut :Integer;
          Options :Integer; Const CommandText :WideString;
          Params :OLEVariant) :OLEVariant; Override;
        Function GetRecordCount :Integer; Override;
        Procedure InternalCancel; Override;
        Procedure InternalClose; Override;
        Procedure InternalDelete; Override;
        Procedure InternalPost; Override;
        Procedure InternalRefresh; Override;
        Procedure Notification (AComponent :TComponent;
          Operation :TOperation); Override;
        Procedure OpenCursor (InfoQuery :Boolean); Override;
        Procedure SetDataSetField (Const Value :TDataSetField); Override;
        Procedure SetFieldData (Field :TField; Buffer :Pointer); Overload;
          Override;

        { Properties }
        Property ParentRecNumber :Cardinal Read GetParentRecNumber;
        Property RecNumber :Cardinal Read GetRecNumber;
        Property StatusFilterAttrs :Integer Read GetStatusFilterAttrs;
        Property WaitingForApplyingConfirmation :Boolean
          Read FWaitingForApplyingConfirmation
          Write SetWaitingForApplyingConfirmation;
      Public
        { Fields }
        ChangeCheckFieldTypes :TghFieldTypes;
        FieldFocused :Boolean;

        Constructor Create (AOwner :TComponent); Override;
        Destructor Destroy; Override;

        { Static methods }
        Function ApplyChangeLog (ConfirmSuccess :TghTernary) :Integer;
          Overload;
        Function ApplyDetailUpdates (MaxErrors :Integer = 0;
          Const ConfirmSuccess :TghTernary = ghtrDefault) :Integer;
        Function ApplyUpdates (Const MaxErrors :Integer;
          Const ConfirmSuccess :TghTernary) :Integer; Reintroduce;
          Overload;
        Function ApplyUpdates (
          Const ConfirmSuccess :TghTernary = ghtrDefault) :Integer;
          Reintroduce; Overload;
        Procedure CancelDetails;
        Procedure CancelDetailUpdates;
        Function CancelRec :TDataSetState;
        Function CloneCursor (Const Reset :Boolean = True;
          Const KeepSettings :Boolean = False) :TghClientDataSet;
          Reintroduce; Overload;
        Function CloneCursorAux (Reset :Boolean = True;
          KeepSettings :Boolean = False) :TghClientDataSet;
        Function ControlsDisabled (
          Const IncludeBlockReadState :Boolean = False) :Boolean;
        Function CreateFieldList (Const Names :String) :TList;
        Function CreateIndexFieldList :TList;
        Function CreateMasterFieldList :TList;
        Procedure Delete;
        Procedure DeleteDetails;
        Procedure DeleteQuiet;
        Procedure DeleteRecs (Const VisibleOnly :Boolean = False;
          Const FilterProc :TghEnumProc = Nil); Overload;
        Procedure DeleteRecs (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant); Overload;
        Procedure DeleteRecs (Const FilterProc :TghEnumProc); Overload;
        Function DetailChangeCount :Integer;
        Function CloneAsDetail (Const AMasterSource :TComponent;
          Const AMasterFields :String; Const AIndexFieldNames :String = '';
          Const AFieldParamNames :String = '') :TghClientDataSet;
        Function DetailCount (Const Update :Boolean = False) :Integer;
        Function DisableControls (Const IncludeNonvisual :Boolean = True) :TBookmark;
        Procedure Empty (Const AProviderEOF :TghTernary = ghtrDefault);
        Procedure EmptyDetails (
          Const AProviderEOF :TghTernary = ghtrDefault);
        Procedure EnableControls (Const ABookmark :TBookmark = Nil);
        Function EnumDataSources (Const Routine :TghEnumProc;
          Const AParams :Variant) :TghDataSource;
        Function GetLogChangesQuiet :Boolean;
        Function GoToBookmarkValid (Const Value :TBookmark) :Boolean;
        Function HasMasterSavePoint :Boolean;
        Function HasVisualLinks :Boolean;
        Function InvalidateSavePoints (Const StartGlobalIndex :Integer = 0;
          Const DetailAction :TghSavePointAction = ghsaNone) :Integer;
        Function LinkDataSource :TghDataSource;
        Function Locate (Const KeyFields :String; Const KeyValues :Variant)
          :Boolean; Reintroduce; Overload;
        Procedure MergeChangeLog;
        Procedure MergeDetailChanges;
        Procedure PostDetails;
        Function QuietDeleting :Boolean;
        Function RecAttrs :Integer;
        Procedure RestoreFilter (Const Filtering :TghDataSetFiltering);
        Procedure RestoreStatusFilter (Const Value :TUpdateStatusSet);
        Procedure RevertChanges (AUpdateStatus :TghChangeUpdateStatusSet;
          Const FilterProc :TghEnumProc; Const FilterParams :Variant); Overload;
        Procedure RevertChanges (
          Const AUpdateStatus :TghChangeUpdateStatusSet = ghusChanges;
          Const FilterProc :TghEnumProc = Nil); Overload;
        Procedure RevertChanges (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant); Overload;
        Procedure RevertChanges (Const FilterProc :TghEnumProc); Overload;
        Procedure RevertDetails;
        Procedure RevertRecord;
        Procedure RevertRecs (Const VisibleOnly :Boolean = False;
          Const FilterProc :TghEnumProc = Nil); Overload;
        Procedure RevertRecs (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant); Overload;
        Procedure RevertRecs (Const FilterProc :TghEnumProc); Overload;
        Function RowCount (Const VisibleOnly :Boolean = False;
          Const FilterProc :TghEnumProc = Nil) :Integer; Overload;
        Function RowCount (Const FilterProc :TghEnumProc;
          Const FilterParams :Variant) :Integer; Overload;
        Function RowCount (Const FilterProc :TghEnumProc) :Integer;
          Overload;
        Function SetLogChangesQuiet (Const Value :Boolean = True) :Boolean;
        Function SetTempFilter (Const Expr :String;
          Const LockNonvisualLinks :Boolean = True) :TghDataSetFiltering;
          Overload;
        Function SetTempStatusFilter (Const Value :TUpdateStatusSet;
          Const LockNonvisualLinks :Boolean = True) :TUpdateStatusSet;

        { Virtual methods }
        Function ApplyChangeLog (MaxErrors :Integer = 0;
          ConfirmSuccess :TghTernary = ghtrDefault) :Integer; Overload;
          Virtual;
        Procedure CancelUpdates; Virtual;
        Procedure DeleteRecs (VisibleOnly :Boolean;
          FilterProc :TghEnumProc; Const FilterParams :Variant); Overload;
          Virtual;
        Function GetFieldParamNames (Const Field :String) :String; Virtual;
        Function FocusField (Const Field :String) :Boolean; Virtual;
        Procedure RevertRecs (VisibleOnly :Boolean;
          FilterProc :TghEnumProc; Const FilterParams :Variant); Overload;
          Virtual;
        Function RowCount (VisibleOnly :Boolean; FilterProc :TghEnumProc;
          Const FilterParams :Variant) :Integer; Overload; Virtual;

        { Overridden methods }
        Function ApplyUpdates (MaxErrors :Integer) :Integer; Overload;
          Override;
        Function BookmarkValid (Bookmark :TBookmark) :Boolean; Override;
        Procedure Cancel; Override;
        Procedure CloneCursor (Source :TCustomClientDataSet;
          Reset :Boolean; KeepSettings :Boolean = False); Overload;
          Override;
        Function DataRequest (Data :OLEVariant) :OLEVariant; Override;
        Procedure GetDetailLinkFields (AMasterFields, DetailFields :TList);
          Override;
        Function GetFieldData (Field :TField; Buffer :Pointer) :Boolean;
          Overload; Override;
        Function Locate (Const KeyFields :String; Const KeyValues :Variant;
          Options :TLocateOptions) :Boolean; Overload; Override;
        Procedure Post; Override;
        Procedure Resync (Mode :TResyncMode); Override;

        { Properties }
        Property Canceling :Boolean Read GetCanceling;
        Property ChangeLog :TghClientDataSet Read GetChangeLog;
        Property Details [Const Index :Integer] :TDataSet Read GetDetail;
        Property Deleting :Boolean Read GetDeleting;
        Property Detailed :Boolean Read GetDetailed;
        Property IsAuxClone :Boolean Read FIsAuxClone;
        Property IsChangeLog :Boolean Read GetIsChangeLog;
        Property Limited :Boolean Read GetLimited;
        Property LogChanges Write SetLogChanges;
        Property Master :TDataSet Read GetMaster;
        Property MasterActive :Boolean Read GetMasterActive;
        Property NestedCount :Integer Read GetNestedCount;
        Property Nesteds [Const Index :Integer] :TghClientDataSet
          Read GetNested;
        Property Parent :TghClientDataSet Read GetParent;
        Property PersistentState :TDataSetState Read FPersistentState;
        Property Posting :Boolean Read GetPosting;
        Property PrevPersistentState :TDataSetState
          Read FPrevPersistentState;
        Property Primary :TghClientDataSet Read GetPrimary;
        Property ProviderEOF;
        Property Ranged :Boolean Read GetRanged;
        Property RootMaster :TghClientDataSet Read GetRootMaster;
        Property RootParent :TghClientDataSet Read GetRootParent;
        Property SortFieldNames :String Read GetSortFieldNames;
        Property StatusFilter Write SetStatusFilter;
        Property UpdateErrors :Integer Read FUpdateErrors;
      Published
        { Properties }
        Property Active Stored FStoreActive;
        Property AllParamsWhenDetail :Boolean Read FAllParamsWhenDetail
          Write FAllParamsWhenDetail Default False;
        Property AutoApplyDetails :Boolean Read GetAutoApplyDetails
          Write FAutoApplyDetails Default True;
        Property AutoCancelDetails :Boolean Read FAutoCancelDetails
          Write FAutoCancelDetails Default True;
        Property AutoDeleteDetails :Boolean Read FAutoDeleteDetails
          Write FAutoDeleteDetails Default False;
        Property AutoEdit :Boolean Read FAutoEdit Write FAutoEdit
          Default True;///que funcione tambin con Blobs
        Property FieldParamNames :String Read FFieldParamNames
          Write FFieldParamNames;
        Property ID :String Read FID Write FID;
        Property Identify :Boolean Read FIdentify Write FIdentify
          Default False;
        Property ImmediateApplyUpdates :Boolean Read FImmediateApplyUpdates
          Write SetImmediateApplyUpdates Default False;
        Property MasterEventOnOpen :Boolean Read FMasterEventOnOpen
          Write FMasterEventOnOpen Default False;
        Property Provider :TCustomProvider Read FProvider
          Write SetProviderProp;
        Property ProviderName Write SetProviderName;
        Property SavePoints :TghClientDataSetSavePoints
          Read GetSavePoints;
        Property StoreActive :Boolean Read FStoreActive Write FStoreActive
          Default False;

        { Event properties }
        Property AfterFieldChange :TghFieldAfterChangeEvent
          Read FAfterFieldChange Write FAfterFieldChange;
        Property BeforeFieldChange :TghFieldBeforeChangeEvent
          Read FBeforeFieldChange Write FBeforeFieldChange;
        Property BeforeFocusField :TghBeforeContinueFieldEvent
          Read FBeforeFocusField Write FBeforeFocusField;
    End;

  Var
    { Client Data Set Base Order }
    GHClientDataSetBaseOrder :String = '_BaseOrder';

    { Client Data Sets }
    GHClientDataSets :TghObjList;

Implementation

  Uses
    SysUtils, Variants, Windows, Math, GHFUtils;

  { TghClientDataSetSavePoint }

  { Protected overridden methods }

  Procedure TghClientDataSetSavePoint.InternalConfirm;
  Begin
    If Index = 0 Then
      With TghClientDataSet (Owner.Owner) Do
      Begin
        InternalMergeChangeLog;
        WaitingForApplyingConfirmation := False;
      End;
  End;

  // Based on TCustomClientDataSet.SetSavePoint method
  Procedure TghClientDataSetSavePoint.InternalRestore;
  Var
    SavedModified :Boolean;
  Begin
    With TghClientDataSet (Owner.Owner) Do
    Begin
      If Not Posting Then
      Begin
        Cancel;
        CheckBrowseMode;
        UpdateCursorPos;
      End;

      SavedModified := Modified;

      Try
        Check (DSBase.RollBack (Self.State));
      Finally
        { Recover Modified property changed by DSBase.RollBack ->
          CalcFieldsCallBack -> SetTempState }
        SetModified (SavedModified);
      End;

      If Not Posting Then
      Begin
        CursorPosChanged;
        Resync ([]);
      End;

      If Index = 0 Then
        WaitingForApplyingConfirmation := False;
    End;
  End;

  // Based on TCustomClientDataSet.GetSavePoint method
  Procedure TghClientDataSetSavePoint.InternalSave;
  Begin
    With TghClientDataSet (Owner.Owner) Do
    Begin
      If Not Posting Then
        CheckBrowseMode;

      Self.State := BaseProp (dspropGetSavePoint);
    End;
  End;

  { Public overridden methods }

  Procedure TghClientDataSetSavePoint.Invalidate (
    DetailAction :TghSavePointAction = ghsaNone);
  Begin
    TghClientDataSet (Owner.Owner).InvalidateNestedSavePoints (
      GlobalIndex);
    Inherited Invalidate (DetailAction);
  End;

  { TghClientDataSetSavePoints }

  { Protected overridden methods }

  Function TghClientDataSetSavePoints.NewPointClass
    :TghCustomSavePointClass;
  Begin
    Result := TghClientDataSetSavePoint;
  End;

  { TghClientDataSet }

  Constructor TghClientDataSet.Create (AOwner :TComponent);
  Begin
    Inherited Create (AOwner);
    AutoApplyDetails := True;
    AutoCancelDetails := True;
    AutoEdit := True;
    ChangeCheckFieldTypes := [ftString, ftSmallint, ftInteger, ftWord,
      ftBoolean, ftFloat, ftCurrency, ftBCD, ftDate, ftTime, ftDateTime,
      ftAutoInc, ftFixedChar, ftWideString, ftLargeint, ftVariant, ftGuid,
      ftTimeStamp, ftFMTBcd];

    If GHClientDataSets = Nil Then
      GHClientDataSets := TghObjList.Create (False);
                                      
    GHClientDataSets.Add (Self);
  End;

  Destructor TghClientDataSet.Destroy;
  Begin
    FreeAndNil (FDetailList);
    FreeAndNil (BlockReadSizeStack);
    FreeAndNil (FSavePoints);
    Inherited Destroy;

    { NOTE: The inherited destructor could call this destructor indirectly
      for another object. } 
    If (GHClientDataSets <> Nil) And (GHClientDataSets.Count = 0) Then
      FreeAndNil (GHClientDataSets);
  End;

  { Protected static methods }

  Function TghClientDataSet.ActiveRecBookmark :PDSBookmrk;
  Begin
    Result := Pointer (ActiveRecBuffer);

    If Result <> Nil Then
      Result := RecBookmark (Result);
  End;

  Function TghClientDataSet.ActiveRecBuffer :PChar;
  Begin
    If State In ghdsCalc Then
      Result := CalcBuffer
    Else
      Result := ActiveBuffer;
  End;

  Function TghClientDataSet.BaseProp (Const Prop :DSProp) :Integer;
  Var
    Props :PDSProps;
  Begin
    { NOTE: IDSBase.GetProp don't work properly with the dspropReadOnly
      constant (DataSnap bug).  It is necessary to use the IDSBase.GetProps
      method in order to get the value of this property. }
    If Prop = dspropReadOnly Then
    Begin
      New (Props);

      Try
        Check (DSBase.GetProps (Props^));
        Result := Integer (Props.bReadOnly);
      Finally
        Dispose (Props);
      End;
    End
    Else
      Check (DSBase.GetProp (Prop, @Result));
  End;

  Function TghClientDataSet.ChangeCurrentRec (Const Cursor :IDSCursor;
    Const AllowCalcs :Boolean) :Boolean;
  Var
    FieldDescs :TFieldDescList;
    I :Integer;
    SavedLogChanges :Boolean;

    Function BlankFlag (Const Field :Integer) :PByte;
    Begin
      // "Blank" flag byte of the given field inside of ActiveBuffer
      Result := ghOffset (ActiveBuffer, FieldDescs [Field].iNullOffsInRec);
    End;
  Begin
    { This auxiliary method allows us to create a new/reversible and
      temporary entry of the current record in the internal record list
      (and in the change log).  If a new entry is created, the old entry is
      made invisible. NOTE: This method was created to solve a native bug
      of the IDSCursor.RevertRecord method (see RevertRecord method in this
      unit) and it assumes that the given cursor is positioned in the
      record that correspond to ActiveBuffer. }

    Result := False;

    If Boolean (BaseProp (dspropReadOnly)) Or
    Boolean (BaseProp (dspropDisableEdits)) Then
      Exit;

    FieldDescs := GetFieldDescs;

    { We ensure that the change will be logged (otherwise the new entry is
      not created) }
    SavedLogChanges := SetLogChangesQuiet;

    { We avoid unnecessary processing (Cursor.ModifyRecord ->
      ChangeNotification -> InternalChangeNotification), because the new
      entry is to be undone (must be undone by the caller). }
    LockChangeNotification;

    Try
      { We search the last field that can to be blanked/un-blanked ("blank"
        flag), in order to create a new (and undoable) entry. NOTE: Changes
        of the blank flag are admitted even in read only fields. }
      For I := High (FieldDescs) DownTo 0 Do
        If Not (FieldDescs [I].iFldType In [fldBlob, fldADT, fldArray,
        fldRef, fldTable]) And
        (FieldDescs [I].iFldAttr In [0, fldAttrReadOnly]) And
        (AllowCalcs Or Not FieldDescs [I].bCalculated) Then
        Begin
          { We increment the blank flag in order to the Cursor.ModifyRecord
            method accepts that the current record changed
            (IDSCursor.ModifyRecord does nothing if the given buffer is
            equal to the internal record) }
          Inc (BlankFlag (I)^);

          Try
            { If the temporary modification can not be done due to a
              conflict of unique key, we will try with the next (I - 1)
              field }
            If CheckStatus (Cursor.ModifyRecord (ActiveBuffer),
            ghdaErrorKeyViol) Then
              Continue;
          Finally
            { We recover the original blank flag of the field in the record
              buffer }
            Dec (BlankFlag (I)^);
          End;

          Result := True;
          Exit;
        End;
    Finally
      UnLockChangeNotification;
      SetLogChangesQuiet (SavedLogChanges);
    End;
  End;

  Function TghClientDataSet.ChangeIndexAttrs :Integer;
  Begin
    Result := BaseProp (dspropChangeIndex_View);
  End;

  Procedure TghClientDataSet.ChangeNotification;
  Var
    SavedInChangeNotification :Boolean;
  Begin
    { We avoid unnecessary/unwanted processing if the data set is updating
      its own cursor, or if the change notification is locked by any of the
      clones }
    If UpdatingCursor Or ChangeNotificationLocked Then
      Exit;

    SavedInChangeNotification := InChangeNotification;
    InChangeNotification := True;

    Try
      Try
        InternalChangeNotification;
      Except
        { Empty Except, like in TCustomClientDataSet.NotifyCallback
          (remember that this is a DataSnap call back routine) }
      End;
    Finally
      InChangeNotification := SavedInChangeNotification;
    End;
  End;

  Function TghClientDataSet.ChangeNotificationLocked :Boolean;
  Var
    I :Integer;
  Begin
    For I := 0 To GHClientDataSets.Count - 1 Do
      If SameBase (I) And (TghClientDataSet (
      GHClientDataSets [I]).ChangeNotificationLockCount > 0) Then
      Begin
        Result := True;
        Exit;
      End;

    Result := False;
  End;

  Procedure TghClientDataSet.CheckDeleteDetails;
  Begin
    If AutoDeleteDetails Then
      DeleteDetails;
  End;

  Procedure TghClientDataSet.CheckIdentify (Var Data :OLEVariant);
  Begin
    If Identify Then
      Data := VarArrayOf ([InternalID (), Data]);
  End;

  Procedure TghClientDataSet.CheckRemoveDetails;
  Begin
    CheckRevertDetails;
    CheckDeleteDetails;
  End;

  Procedure TghClientDataSet.CheckRevertDetails;
  Begin
    If AutoCancelDetails Then
      RevertDetails;
  End;

  Function TghClientDataSet.CheckStatus (Const Status, Value :DBResult)
    :Boolean;
  Begin
    If Status = Value Then
      Result := True
    Else
    Begin
      Check (Status);  // Exception if Status <> 0
      Result := False;  // Status = 0
    End;
  End;

  Procedure TghClientDataSet.CheckWaitingForApplyingConfirmation;
  Begin
    If RootParent.WaitingForApplyingConfirmation Then
      Raise EInvalidOperation.Create (ghFormatOptional (
        'Invalid Operation.  Data set%*s is waiting for confirmation of ' +
        'applied updates.', Name));
  End;

  Function TghClientDataSet.Contains (DataSet :TghClientDataSet)
    :Boolean;
  Begin
    While (DataSet <> Self) And (DataSet.Parent <> Nil) Do
      DataSet := DataSet.Parent;

    Result := DataSet = Self;
  End;

  Function TghClientDataSet.CreateCursor (
    Const ABookmark :TBookmark = Nil) :IDSCursor;
  Begin
    CreateDbClientObject (ClsID_DSCursor, IDSCursor, Result);
    Result.InitCursor (DSBase);
    RegChangeNotification;

    If ABookmark <> Nil Then                   
      Check (Result.MoveToBookMark (ABookmark));
  End;

  Function TghClientDataSet.CurrentBookmark :DSBookmrk;
  Begin
    Check (DSCursor.GetCurrentBookMark (@Result));
  End;

  Function TghClientDataSet.CursorRecAttrs (Const Cursor :IDSCursor)
    :DSAttr;
  Begin
    Check (Cursor.GetRecordAttribute (Result));
  End;

  Function TghClientDataSet.CursorRecCount (Const Cursor :IDSCursor)
    :Integer;
  Begin
    If CheckStatus (Cursor.GetRecordCount (Result),
    dberr_DetailsNotFetched) Then
      Result := 0;
  End;

  Function TghClientDataSet.CursorRecNumber (Const Cursor :IDSCursor)
    :Cardinal;
  Begin
    Check (Cursor.GetRecordNumber (Result));
  End;

  Procedure TghClientDataSet.DeleteRecsDetail (Const AMaster :TDataSet;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    DoRecsDetail (AMaster, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.DeleteRecsFilter)), FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.DeleteRecsFilter (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Var
    Clone :TghClientDataSet;
  Begin
    // Auxiliary clean clone (not filtered or ranged)
    Clone := CloneCursorAux;
    Clone.AutoCancelDetails := AutoCancelDetails;
    Clone.AutoDeleteDetails := AutoDeleteDetails;

    Try
      Clone.DeleteRecsVisible (FilterProc, FilterParams);
    Finally
      Clone.Free;
    End;
  End;

  {}//probar
  Procedure TghClientDataSet.DeleteRecsNested (Const ParentRec :Cardinal;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    DoRecsNested (ParentRec, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.DeleteRecsFilter)), FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.DeleteRecsVisible (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Var
    ARecNo :Integer;
  Begin
    First;

    While Not EOF Do
    Begin
      If Not Assigned (FilterProc) Or
      ghCall (FilterProc, Self, FilterParams) Then
      Begin
        ARecNo := RecNo;
        Delete;

        If RecNo = ARecNo Then  // If the removed row was not the last one
          Continue;
      End;

      Next;
    End;
  End;

  {$Warn No_RetVal Off}
  Function TghClientDataSet.DetailList (Const Update :Boolean = False)
    :TObjectList;
  Var
    I :Integer;
  Begin
    Try
      If FDetailList = Nil Then
        FDetailList := TObjectList.Create (False)
      Else
        If Update Then
          For I := 0 To FDetailList.Count - 1 Do
            TDataSet (FDetailList [I]).RemoveFreeNotification (Self)
        Else
          Exit;

      { NOTE: The TDataSet.GetDetailDataSets method returns this list in
        reverse order of data set binding.  We do not override
        GetDetailDataSets because that is considered normal.  In contrast,
        the new Details property returns the detail data sets as they were
        bound (which generally corresponds to creation order). }
      GetDetailDataSets (FDetailList);
      ghReverse (FDetailList);

      For I := 0 To FDetailList.Count - 1 Do
        TDataSet (FDetailList [I]).FreeNotification (Self);
    Finally
      Result := FDetailList;
    End;
  End;
  {$Warn No_RetVal On}

  Function TghClientDataSet.FieldDesc (Const Field :String) :DSFLDDesc;
  Var
    FieldDescs :TFieldDescList;
    I :Integer;
  Begin
    FieldDescs := GetFieldDescs;
    
    For I := 0 To High (FieldDescs) Do
      If ghEquals (FieldDescs [I].szName, Field) Then
      Begin
        Result := FieldDescs [I];
        Exit;
      End;

    ZeroMemory (@Result, SizeOf (Result));
  End;

  Function TghClientDataSet.FindSavePoints
    :TghClientDataSetSavePoints;
  Var
    I :Integer;
  Begin
    If FSavePoints = Nil Then
      For I := 0 To GHClientDataSets.Count - 1 Do
        If SameBase (I) And
        (TghClientDataSet (GHClientDataSets [I]).FSavePoints <> Nil) Then
        Begin
          Result := TghClientDataSet (GHClientDataSets [I]).FSavePoints;
          Exit;
        End;

    Result := FSavePoints
  End;

  Function TghClientDataSet.GetAutoApplyDetails :Boolean;
  Var
    DataSet :TghClientDataSet;
  Begin
    DataSet := Self;

    Repeat
      Result := DataSet.FAutoApplyDetails;

      If Result Or (DataSet.Parent = Nil) Then
        Exit
      Else
        DataSet := DataSet.Parent;
    Until False;
  End;

  Function TghClientDataSet.GetCanceling :Boolean;
  Var
    DataSet :TghClientDataSet;
  Begin
    DataSet := Self;

    Repeat
      Result := DataSet.FCanceling;

      If Result Or (DataSet.Parent = Nil) Then
        Exit
      Else
        DataSet := DataSet.Parent;
    Until False;
  End;

  Function TghClientDataSet.GetDeleting :Boolean;
  Var
    DataSet :TghClientDataSet;
  Begin
    DataSet := Self;

    Repeat
      Result := DataSet.FDeleting;

      If Result Or (DataSet.Parent = Nil) Then
        Exit
      Else
        DataSet := DataSet.Parent;
    Until False;
  End;

  Function TghClientDataSet.GetDetail (Const Index :Integer) :TDataSet;
  Begin
    Result := DetailList [Index] As TDataSet;
  End;

  Function TghClientDataSet.GetDetailed :Boolean;
  Begin
    Result := (DataSetField <> Nil) Or MasterActive;
  End;

  Function TghClientDataSet.GetFieldData (Const Cursor :IDSCursor;
    Const RecBuffer :Pointer; Const Field :Integer;
    Const Data :Pointer = Nil) :Boolean;
  Var
    Blank :LongBool;
  Begin
    Check (Cursor.GetField (RecBuffer, Field, Data, Blank));
    Result := Not Blank;
  End;

  Function TghClientDataSet.GetFieldData (Const RecBuffer :Pointer;
    Const Field :Integer; Const Data :Pointer = Nil) :Boolean;
  Begin
    Result := GetFieldData (DSCursor, RecBuffer, Field, Data);
  End;

  // Based on TCustomClientDataSet.InternalInitFieldDefs method
  Function TghClientDataSet.GetFieldDescs :TFieldDescList;
  Var
    Props :DSProps;
  Begin
    DSBase.SetProp (dspropCompressArrays, 1);
    DSBase.GetProps (Props);
    SetLength (Result, Props.iFields);
    DSBase.GetFieldDescs (PDSFldDesc (Result));
  End;

  Procedure TghClientDataSet.GetIndexFieldList (Const List :TList);
  Var
    I :Integer;
  Begin
    List.Clear;

    For I := 0 To IndexFieldCount - 1 Do
      List.Add (IndexFields [I]);
  End;

  Function TghClientDataSet.GetIsChangeLog :Boolean;
  Begin
    Result := IsAuxClone And (StatusFilter <> []);
  End;

  Function TghClientDataSet.GetLimited :Boolean;
  Begin
    Result := Filtered Or Ranged;
  End;

  Function TghClientDataSet.GetMaster :TDataSet;
  Begin
    If MasterSource <> Nil Then
      Result := MasterSource.DataSet
    Else
      Result := Nil;
  End;

  Function TghClientDataSet.GetMasterActive :Boolean;
  Begin
    Result := (MasterSource <> Nil) And (MasterSource.State <> dsInactive);
  End;

  Procedure TghClientDataSet.GetMasterFieldList (Const List :TList);
  Begin
    List.Clear;

    If Master <> Nil Then
      Master.GetFieldList (List, MasterFields);
  End;

  Function TghClientDataSet.GetNested (Const Index :Integer)
    :TghClientDataSet;
  Begin
    Result := TDataSet (NestedDataSets [Index]) As TghClientDataSet;
  End;

  Function TghClientDataSet.GetNestedCount :Integer;
  Begin
    Result := NestedDataSets.Count;
  End;

  Function TghClientDataSet.GetParent :TghClientDataSet;
  Begin
    If DataSetField <> Nil Then
      Result := DataSetField.DataSet As TghClientDataSet
    Else
      Result := Nil;
  End;

  Function TghClientDataSet.GetParentRecNumber :Cardinal;
  Var
    Blank :LongBool;
  Begin
    If (DataSetField <> Nil) And Not IsEmpty Then
      Check (DSCursor.GetField (ActiveBuffer,
        FieldDesc ('Link_Field').iFieldID, @Result, Blank))
    Else
      Result := 0;
  End;

  Function TghClientDataSet.GetPosting :Boolean;
  Var
    DataSet :TghClientDataSet;
  Begin
    DataSet := Self;

    Repeat
      Result := DataSet.FPosting;

      If Result Or (DataSet.Parent = Nil) Then
        Exit
      Else
        DataSet := DataSet.Parent;
    Until False;
  End;

  Function TghClientDataSet.GetPrimary :TghClientDataSet;
  Begin
    If IsAuxClone Then
      Result := (CloneSource As TghClientDataSet).Primary
    Else
      Result := Self;
  End;

  Function TghClientDataSet.GetRanged :Boolean;
  Begin
    Result := GetKeyBuffer (kiCurRangeStart).Modified Or
      GetKeyBuffer (kiCurRangeEnd).Modified;
  End;

  Function TghClientDataSet.GetRecNumber :Cardinal;
  Begin
    UpdateCursorPos;
    Result := InternalGetRecNumber;
  End;

  Function TghClientDataSet.GetRootMaster :TghClientDataSet;
  Begin
    Result := RootParent;

    If Result.Master = Nil Then
      Exit;

    While Result.Master <> Nil Do
      Result := Result.Master As TghClientDataSet;

    Result := Result.RootMaster;
  End;

  Function TghClientDataSet.GetRootParent :TghClientDataSet;
  Begin
    Result := Self;

    While Result.Parent <> Nil Do
      Result := Result.Parent;
  End;

  Function TghClientDataSet.GetSavePoints
    :TghClientDataSetSavePoints;
  Begin
    Result := FindSavePoints;

    If Result <> Nil Then
      Exit;

    FSavePoints := SavePointsClass.Create (Self);
    ghSetSubcomponent (FSavePoints, 'SavePoints');
    Result := FSavePoints;
  End;

  Function TghClientDataSet.GetSortFieldNames :String;
  Var
    I :Integer;
  Begin
    If IndexFieldCount > 0 Then
    Begin
      Result := '';

      For I := 0 To IndexFieldCount - 1 Do
        Result := ghSuffixed (Result, ';') + IndexFields [I].FieldName;
    End
    Else  // Maybe data set is closed
    Begin
      Result := IndexFieldNames;

      If (Result = '') And (IndexName <> '') Then
        Result := IndexDefs.Find (IndexName).Fields;
    End;
  End;

  Function TghClientDataSet.GetStatusFilterAttrs :Integer;
  Begin
    Result := ghDataSnapAttrs (StatusFilter);
  End;

  Function TghClientDataSet.HasSavePoint (
    Const SavePoint :TghCustomSavePoint) :Boolean;
  Begin
    Result := (SavePoint Is TghClientDataSetSavePoint) And
      Contains (TghClientDataSet (SavePoint.Owner.Owner));
  End;

  Function TghClientDataSet.InternalGetRecNumber :Cardinal;
  Begin
    Result := CursorRecNumber (DSCursor);
  End;

  Function TghClientDataSet.InternalID :WideString;
  Begin
    If ID <> '' Then
      Result := ID
    Else
    Begin
      Result := ghGUIDStr ();

      If Not (csDesigning In ComponentState) Then
        ID := Result;
    End;
  End;

  Function TghClientDataSet.InvalidateNestedSavePoints (
    Const StartGlobalIndex :Integer = 0) :Integer;
  Var
    I :Integer;
  Begin
    Result := -1;

    For I := 0 To NestedCount - 1 Do
      Result := ghMinCardinal (Result, Nesteds [I].InvalidateSavePoints (
        StartGlobalIndex));
  End;

  Procedure TghClientDataSet.LockChangeNotification;
  Begin
    Inc (ChangeNotificationLockCount);
  End;

  Procedure TghClientDataSet.LockResync;
  Begin
    Inc (ResyncLockCount);
  End;

  Function TghClientDataSet.NativeValue (Const Field :TField;
    Const Buffer :Pointer) :Variant;
  Begin
    TempFieldData := @Buffer;

    Try
      Result := Field.Value;
    Finally
      TempFieldData := Nil;
    End;
  End;

  Function TghClientDataSet.RecBookmark (Const RecBuffer :Pointer)
    :PDSBookmrk;
  Begin
    Result := ghOffset (RecBuffer, RecordSize + CalcFieldsSize +
      SizeOf (TRecInfo));
  End;

  Function TghClientDataSet.RecInfo (Const RecBuffer :Pointer)
    :PRecInfo;
  Begin
    Result := ghOffset (RecBuffer, RecordSize + CalcFieldsSize);
  End;

  {}//probar
  Procedure TghClientDataSet.RefreshChangeLogs (
    Const IncludeSelf :Boolean = True);
  Var
    ChangeIndexUpdated :Boolean;

    // Call back routine
    Procedure DoSameBase (Const DataSet :TghClientDataSet;
      Const AParams :Variant; Var Continue :Boolean);
    Var
      ASelf :TghClientDataSet;
      AChangeIndexUpdated :PBoolean;
    Begin
      ASelf := TghClientDataSet (ghPtr (AParams [0]));
      AChangeIndexUpdated := ghPtr (AParams [2]);

      If (DataSet.DSBase = ASelf.DSBase) And (AParams [1] Or
      (DataSet <> ASelf)) And (DataSet.StatusFilter <> []) Then
      Begin
        If Not AChangeIndexUpdated^ Then
        Begin
          DataSet.InternalRefresh;
          AChangeIndexUpdated^ := True;
        End;

        DataSet.Resync ([]);
      End;
    End;
  Begin
    ChangeIndexUpdated := False;
    GHClientDataSets.Enum (@DoSameBase, VarArrayOf (
      [ghVar (Self), IncludeSelf, ghVar (@ChangeIndexUpdated)]));
  End;

  Procedure TghClientDataSet.RegChangeNotification;
  Begin
    If ChangeNotificationReg Then
      Exit;

    Check (DSCursor.SetNotifyCallBack (Integer (Self),
      @TghClientDataSet.ChangeNotification));
    ChangeNotificationReg := True;
  End;

  Function TghClientDataSet.ResyncLocked :Boolean;
  Begin
    Result := ResyncLockCount > 0;
  End;

  Procedure TghClientDataSet.RevertRecsDetail (Const AMaster :TDataSet;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    DoRecsDetail (AMaster, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.RevertRecsFilter)), FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.RevertRecsFilter (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Var
    Clone :TghClientDataSet;
  Begin
    If AutoCancelDetails Then
    Begin
      // Auxiliary clean clone (not filtered or ranged)
      Clone := CloneCursorAux;
      Clone.AutoDeleteDetails := AutoDeleteDetails;

      Try
        { We revert the visible rows first and then the deleted ones, in
          order to avoid the "key violation" error due to new records with
          the same key.  After reverting visible rows, RevertChanges will
          revert the remaining changes (unmodified, modified or inserted
          records that were deleted). }
        Clone.RevertRecsVisible (FilterProc, FilterParams);
      Finally
        Clone.Free;
      End;
    End;

    { If AutoCancelDetails is False, we revert change rows only. That is,
      it makes not sense to traverse through all the records if we do not
      have to revert detail rows. }
    RevertChanges (FilterProc, FilterParams);
  End;

  {}//probar
  Procedure TghClientDataSet.RevertRecsNested (
    Const ParentRec :Cardinal; Const FilterProc :TghEnumProc;
    Const FilterParams :Variant);
  Begin
    DoRecsNested (ParentRec, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.RevertRecsFilter)), FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.RevertRecsVisible (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Var
    ARecNo, ARecordCount :Integer;
  Begin
    { It makes sense to begin by the end because the StatusFilter property
      could be a non-empty set; thus, we first take the last version of
      each record. }
    Last;

    While Not BOF Do
    Begin
      If Not Assigned (FilterProc) Or
      ghCall (FilterProc, Self, FilterParams) Then
      Begin
        ARecNo := RecNo;
        ARecordCount := RecordCount;
        RevertRecord;

        If IsEmpty Or (ARecNo = 1) Then  // No more rows to check
          Exit
        Else
          // If one or two rows were removed
          If RecordCount < ARecordCount Then
          Begin
            RecNo := Min (ARecNo - 1, RecordCount);
            Continue;
          End;
      End;

      Prior;
    End;
  End;

  Function TghClientDataSet.RowCountDetail (Const AMaster :TDataSet;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant)
    :Integer;
  Begin
    Result := DoRecsDetail (AMaster, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.RowCountFilter)), FilterProc, FilterParams);
  End;

  Function TghClientDataSet.RowCountFilter (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
  Begin
    If Assigned (FilterProc) Then
      // Auxiliary clean clone (not filtered or ranged)
      With CloneCursorAux Do
        Try
          Result := RowCountVisible (FilterProc, FilterParams);
        Finally
          Free;
        End
    Else
      If Limited Then
        // Auxiliary clean cursor (not filtered or ranged)
        Result := CursorRecCount (CreateCursor)
      Else
        Result := RecordCount;
  End;

  {}//probar
  Function TghClientDataSet.RowCountNested (Const ParentRec :Cardinal;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
  Begin
    Result := DoRecsNested (ParentRec, TghDoEnumMethod (ghMethodRec (Self,
      @TghClientDataSet.RowCountFilter)), FilterProc, FilterParams);
  End;

  Function TghClientDataSet.RowCountVisible (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
  Begin
    If Assigned (FilterProc) Then
    Begin
      Result := 0;
      First;

      While Not EOF Do
      Begin
        If ghCall (FilterProc, Self, FilterParams) Then
          Inc (Result);

        Next;
      End;
    End
    Else
      Result := RecordCount
  End;

  Function TghClientDataSet.SameBase (Const DataSetIndex :Integer)
    :Boolean;
  Begin
    Result := TghClientDataSet (GHClientDataSets [DataSetIndex]).DSBase =
      DSBase;
  End;

  Procedure TghClientDataSet.SetBaseOrder (Const Cursor :IDSCursor);
  Var
    Index :DSIDXDesc;

    Function Activate :DBResult;
    Begin
      Result := Cursor.UseIndexOrder (PChar (GHClientDataSetBaseOrder));
    End;
  Begin
    If Not CheckStatus (Activate, dberr_NoSuchIndex) Then
      Exit;

    { We add the "natural" base index (rows as they were created/appended).
      NOTE: the default order (szDefault_Order) is not necessarily equal to
      this "base order". }
    ZeroMemory (@Index, SizeOf (Index));
    StrLCopy (Index.szName, PChar (GHClientDataSetBaseOrder),
      SizeOf (Index.szName) - 1);
    Check (DSBase.CreateIndex (Index));

    Check (Activate);
  End;

  Procedure TghClientDataSet.SetFieldData (Const Cursor :IDSCursor;
    Const RecBuffer :Pointer; Const Field :Integer;
    Const Data :Pointer = Nil);
  Begin
    Check (Cursor.PutField (RecBuffer, Field, Data));
  End;

  Procedure TghClientDataSet.SetFieldData (Const RecBuffer :Pointer;
    Const Field :Integer; Const Data :Pointer = Nil);
  Begin
    SetFieldData (DSCursor, RecBuffer, Field, Data);
  End;

  Procedure TghClientDataSet.SetImmediateApplyUpdates (
    Const Value :Boolean);
  Begin
    If Value = FImmediateApplyUpdates Then
      Exit;

    FImmediateApplyUpdates := Value And
      ((ChangeCount = 0) Or (ApplyUpdates () = 0));

    // Mutually exclusive properties
    If ImmediateApplyUpdates Then
      DataSetField := Nil;
  End;

  Procedure TghClientDataSet.SetInternalBlankFlag (
    Const Rec, Field :Integer; Const Value :Byte);
  Begin
    Check (DSBase.PutBlank (Nil, Rec, Field, Value));
  End;

  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.SetLogChanges (Const Value :Boolean);
  Begin
    Inherited LogChanges := Value;

    { We recover the visibility of the change log; this is lost when the
      LogChanges property is set to False and then some change occurs in
      the data (possible DataSnap bug) }
    If Value Then
      RefreshChangeLogs;
  End;

  Procedure TghClientDataSet.SetProviderName (Const Value :String);
  Begin
    Inherited ProviderName := Value;

    If (Provider <> Nil) And (Value <> Provider.Name) Then
      Provider := Nil;
  End;

  Procedure TghClientDataSet.SetProviderProp (
    Const Value :TCustomProvider);
  Begin
    If Value = Provider Then
      Exit;

    FProvider := Value;

    If Value <> Nil Then
    Begin
      Value.FreeNotification (Self);
      Inherited ProviderName := Value.Name;
    End;

    SetProvider (Value);
  End;

  Function TghClientDataSet.SetUpdatingCursor (
    Const Value :Boolean = True) :Boolean;
  Begin
    Result := UpdatingCursor;
    UpdatingCursor := Value;
  End;

  Procedure TghClientDataSet.SetWaitingForApplyingConfirmation (
    Const Value :Boolean);
  Var
    I :Integer;
  Begin
    For I := 0 To GHClientDataSets.Count - 1 Do
      If SameBase (I) Then
        TghClientDataSet (
          GHClientDataSets [I]).FWaitingForApplyingConfirmation := Value;
  End;

  Procedure TghClientDataSet.UnLockChangeNotification;
  Begin
    Dec (ChangeNotificationLockCount);
  End;

  Procedure TghClientDataSet.UnLockResync;
  Begin
    Dec (ResyncLockCount);
  End;

  Procedure TghClientDataSet.UpdateChangeIndex (Attrs :Integer = -1);
  Begin
    If Attrs = -1 Then
      Attrs := ChangeIndexAttrs;  // Current attributes of visibility

    If Attrs = 0 Then
      Attrs := ghdaRecAttrsChanges;

    // This recreates the szChangeIndex index
    Check (DSBase.SetProp (dspropChangeIndex_View, 0));
    Check (DSBase.SetProp (dspropChangeIndex_View, Attrs));
  End;

  { Protected virtual methods }

  Function TghClientDataSet.CanCheckFieldChange (Field :TField)
    :Boolean;
  Begin
    Result := Field.DataType In ChangeCheckFieldTypes;
  End;

  Procedure TghClientDataSet.DoAfterFieldChange (Field :TField;
    Const PrevValue :Variant);
  Begin
    If Assigned (AfterFieldChange) Then
      AfterFieldChange (Field, PrevValue);
  End;

  Procedure TghClientDataSet.DoBeforeFieldChange (Field :TField;
    Const NewValue :Variant);
  Begin
    If Assigned (BeforeFieldChange) Then
      BeforeFieldChange (Field, NewValue);
  End;

  Function TghClientDataSet.DoRecsDetail (AMaster :TDataSet;
    Method :TghDoEnumMethod; FilterProc :TghEnumProc;
    Const FilterParams :Variant) :Integer;
  Var
    AMasterFields :TList;
    DetailFields :TComponentList;

    // Call back routine
    Procedure VerifyRec (Const DataSet :TghClientDataSet;
      Const AParams :Variant; Var Accept :Boolean);
    Var
      AMasterFields :TList;
      DetailFields :TComponentList;
    Begin
      { NOTE: DataSet can be an auxiliary clone given by the operation's
        method (Method).  DetailFields will be emptied if DataSet (and its
        fields) is destroyed, and then this list can be filled with fields
        from another data set (RevertRecsFilter, for example, uses two
        data sets). }

      { AParams [0] = FilterProc
        AParams [1] = FilterParams address
        AParams [2] = Outer AMasterFields
        AParams [3] = Outer DetailFields
        AParams [4] = SortFieldNames
        AParams [5] = DoRecsDetail's Result address }

      AMasterFields := TList (ghPtr (AParams [2]));
      DetailFields := TComponentList (ghPtr (AParams [3]));

      If (DetailFields.Count = 0) Or
      (TField (DetailFields [0]).DataSet <> DataSet) Then
      Begin
        DetailFields.Clear;
        DataSet.GetFieldList (DetailFields, AParams [4]);
        DetailFields.Count := AMasterFields.Count;
      End;

      Accept := ghEqualFields (AMasterFields, DetailFields) And
        ghCallNestedEnumProc (AParams, DataSet, True);

      If Accept Then
        Inc (PInteger (ghPtr (AParams [5]))^);
    End;
  Begin
    AMasterFields := TList.Create;

    Try
      AMaster.GetFieldList (AMasterFields, MasterFields);
      DetailFields := TComponentList.Create (False);

      Try
        Result := 0;
        Method (@VerifyRec, ghVar ([@FilterProc, @FilterParams,
          AMasterFields, DetailFields, SortFieldNames, @Result]));
      Finally
        DetailFields.Free;
      End;
    Finally
      AMasterFields.Free;
    End;
  End;

  {}//probar
  Function TghClientDataSet.DoRecsNested (ParentRec :Cardinal;
    Method :TghDoEnumMethod; FilterProc :TghEnumProc;
    Const FilterParams :Variant) :Integer;

    // Call back routine
    Procedure VerifyRec (Const DataSet :TghClientDataSet;
      Const AParams :Variant; Var Accept :Boolean);
    Begin
      { NOTE: DataSet can be an auxiliary clone given by the operation's
        method (Method). }

      { AParams^ [0] = FilterProc
        AParams^ [1] = FilterParams address
        AParams^ [2] = ParentRec
        AParams^ [3] = DoRecsNested's Result address }

      Accept := (DataSet.ParentRecNumber = AParams [2]) And
        ghCallNestedEnumProc (AParams, DataSet, True);

      If Accept Then
        Inc (PInteger (ghPtr (AParams [3]))^);
    End;
  Begin
    Result := 0;
    Method (@VerifyRec, ghVar ([@FilterProc, @FilterParams, ParentRec,
      @Result]));
  End;

  {}//probar
  Function TghClientDataSet.GetChangeLog :TghClientDataSet;
  Var
    I :Integer;
  Begin
    If IsChangeLog Then
      Result := Self
    Else
    Begin
      For I := 0 To GHClientDataSets.Count - 1 Do
        If SameBase (I) And TghClientDataSet (
        GHClientDataSets [I]).IsChangeLog And (TghClientDataSet (
        GHClientDataSets [I]).CloneSource = Primary) Then
        Begin
          Result := TghClientDataSet (GHClientDataSets [I]);
          Exit;
        End;

      Result := Primary.CloneCursorAux;
      I := ChangeIndexAttrs;

      If I = 0 Then
        Result.StatusFilter := ghusChanges
      Else
        Result.StatusFilter := ghDataSnapUpdateStatusSet (I);
    End;
  End;

  { Based on TCustomClientDataSet.NotifyCallback method (see also the
    ChangeNotification method) }
  Procedure TghClientDataSet.InternalChangeNotification;
  Begin
    If State In ghdsBrowsing Then
    Begin
      UpdateCursorPos;

      If StatusFilter <> [] Then
        InternalRefresh;

      Resync ([]);
      UpdateCursorPos;
    End;
  End;

  // Based on inherited MergeChangeLog method
  Procedure TghClientDataSet.InternalMergeChangeLog;
  Var
    AChangeCount :Integer;
  Begin
    If Not (Posting Or InternalDeleting) Then
      CheckBrowseMode;

    AChangeCount := ChangeCount;

    { Even if ChangeCount is 0, we call DSBase.AcceptChanges method to
      avoid "key violation" errors with internally calculated fields
      (possible DataSnap bug) }
    DSBase.AcceptChanges;

    If (AChangeCount > 0) And Not Posting And Not InternalDeleting Then
    Begin
      UpdateCursorPos;
      Resync ([]);
    End;

    RefreshChangeLogs (False);

    If AutoApplyDetails Then
      MergeDetailChanges;
  End;

  Function TghClientDataSet.LinkDataSourceClass :TghDataSourceClass;
  Begin
    Result := TghDataSource;
  End;

  Function TghClientDataSet.MustNotifyFieldChange (Field :TField)
    :Boolean;
  Type
    TChangeMethod = Procedure (Field :TField; Const Value :Variant) Of
      Object;
  Var
    ChangeMethod :TMethod;
  Begin
    { We check if the AfterFieldChange / BeforeFieldChange event was
      assigned or the DoAfterFieldChange / DoBeforeFieldChange method was
      overridden }
    If Assigned (AfterFieldChange) Or Assigned (BeforeFieldChange) Then
      Result := True
    Else
    Begin
      TChangeMethod (ChangeMethod) := DoAfterFieldChange;

      If ChangeMethod.Code <> @TghClientDataSet.DoAfterFieldChange Then
        Result := True
      Else
      Begin
        TChangeMethod (ChangeMethod) := DoBeforeFieldChange;
        Result := ChangeMethod.Code <>
          @TghClientDataSet.DoBeforeFieldChange;
      End;
    End;
  End;

  Function TghClientDataSet.SavePointsClass
    :TghClientDataSetSavePointsClass;
  Begin
    Result := TghClientDataSetSavePoints;
  End;

  {}//probar
  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.SetStatusFilter (
    Const Value :TUpdateStatusSet);

    // Call back routine
    Procedure DoSameBase (Const DataSet :TghClientDataSet;
      Const AParams :Variant; Var Continue :Boolean);
    Var
      ASelf :TghClientDataSet;
      AValue :TUpdateStatusSet;
    Begin
      ASelf := TghClientDataSet (ghPtr (AParams [0]));
      AValue := TUpdateStatusSet (Byte (AParams [1]));

      If (DataSet.DSBase = ASelf.DSBase) And
      (DataSet.StatusFilter - AValue <> []) Then
        DataSet.StatusFilter := AValue;
    End;
  Begin
    Inherited StatusFilter := Value;

    { There are inconsistencies if two cursors share the same data and both
      use their StatusFilter property with different values
      (dspropChangeIndex_View is a property in IDSBase, not in IDSCursor).
      If this data set is using its StatusFilter property, we give the same
      value to the StatusFilter property (provided it is a non-empty set)
      of all clones. }
    If Value <> [] Then
      GHClientDataSets.Enum (@DoSameBase,
        VarArrayOf ([ghVar (Self), Byte (Value)]));
  End;

  { Protected overridden methods }

  Procedure TghClientDataSet.AddDataPacket (Const Data :OLEVariant;
    HitEOF :Boolean);
  Var
    AuxCursor :IDSCursor;
    FieldDescs :TFieldDescList;
    FirstRec, I, J :Integer;
    SavedUpdatingCursor :Boolean Absolute FirstRec;
  Begin
    { NOTE: The IDSBase.AppendData method does not copy internally
      calculated fields, but, if called multiple times, it sets their
      "blank" flag incorrectly (DataSnap bug).  We correct this problem,
      by activating that flag on the added records. }

    { We need a clean cursor (not filtered or ranged) in "base order",
      with the purpose of to get the internal record number of the first
      row added }
    AuxCursor := CreateCursor;
    SetBaseOrder (AuxCursor);
    Check (AuxCursor.MoveToEOF);

    // Current last record (or BOF if the cursor is empty)
    CheckStatus (AuxCursor.MoveRelative (-1), dberr_BOF);

    SavedUpdatingCursor := SetUpdatingCursor;

    Try
      // We append the row packet to internal record list
      Inherited AddDataPacket (Data, HitEOF);
    Finally
      UpdatingCursor := SavedUpdatingCursor;
    End;

    // Go to first record added (if there is one)
    If Not CheckStatus (AuxCursor.MoveRelative (1), dberr_EOF) Then
    Begin
      FirstRec := CursorRecNumber (AuxCursor);
      FieldDescs := GetFieldDescs;  // Field descriptors

      For I := 0 To High (FieldDescs) Do
        If FieldDescs [I].bCalculated Then  // InternalCalc field
          { From the first record added up to the last one + 1 (the limit
            for the IDSBase.PutBlank method is the internal "iRecNoNext"
            index), we put the field to "blank" by simply activating its
            blank flag in each record }
          For J := FirstRec To MaxInt Do
            If CheckStatus (DSBase.PutBlank (Nil, J,
            FieldDescs [I].iFieldID, 1), ghdaErrorOutOfRange) Then
              Break;
    End;

    { We recover the visibility of the change log, this is lost when the
      IDSBase.AppendData method is called (possible DataSnap bug) }
    RefreshChangeLogs;
  End;

  Procedure TghClientDataSet.CheckDetailRecords;
  Var
    SavedStatusFilter :TUpdateStatusSet;
  Begin
    CheckingDetailRecs := True;

    Try
      If DataSetField <> Nil Then
        Inherited CheckDetailRecords
      Else
        { We avoid a second call to DoGetRecords on opening (see also the
          TghClientDataSet.DoGetRecords method) }
        If Not Active Then
          Exit
        Else
          { We avoid refetching hidden (due to a filter), deleted or
            modified records }
          If (RecordCount = 0) And (RowCount = 0) Then  // No hidden
          Begin
            SavedStatusFilter := ChangeLog.SetTempStatusFilter (ghusChanges);

            Try
              // No deleted/modified rows related to the same master record
              If DoRecsDetail (Master, TghDoEnumMethod (ghMethodRec (
              ChangeLog, @TghClientDataSet.RowCountVisible)), Nil, Null) =
              0 Then
                Inherited CheckDetailRecords;  // OK, we fetch the records
            Finally
              ChangeLog.RestoreStatusFilter (SavedStatusFilter);
            End;
          End;
    Finally
      CheckingDetailRecs := False;
    End;
  End;

  {}//probar
  Procedure TghClientDataSet.CloseCursor;

    // Call back routine
    Procedure DoSameBase (Const DataSet :TghClientDataSet;
      Const AParams :Variant; Var Continue :Boolean);
    Var
      ASelf :TghClientDataSet;
    Begin
      ASelf := TghClientDataSet (ghPtr (AParams));

      If (DataSet.DSBase = ASelf.DSBase) And DataSet.IsAuxClone And
      (DataSet.Owner = ASelf) Then
        DataSet.Free;
    End;
  Begin
    { We destroy the auxiliary clones that are still owned for this data
      set (the application can change the owner of the clone in order to it
      is not destroyed here) }
    GHClientDataSets.Enum (@DoSameBase, ghVar (Self));

    Inherited CloseCursor;
  End;

  Procedure TghClientDataSet.DataEvent (Event :TDataEvent;
    Info :Longint);
  Begin
    If Event = deUpdateState Then
    Begin
      FPrevPersistentState := PersistentState;
      FPersistentState := State;
      UpdatingState := True;///quiz mejor propiedad indexada Boolean usando contadores de eventos
    End
    Else
      If Event = deFocusControl Then
        FieldFocused := False;

    Try
      Inherited DataEvent (Event, Info);
    Finally
      UpdatingState := False;
    End;

    If (Event = deFocusControl) And (TFieldRef (Info)^ = Nil) Then
      FieldFocused := True;
  End;

  Procedure TghClientDataSet.DoAfterDelete;
  Begin
    FDeleting := False;

    If Not QuietDeleting Then
      Inherited DoAfterDelete;
  End;

    Type
      TDataSetAccess = Class (TDataSet);
  Procedure TghClientDataSet.DoAfterOpen;
  Var
    MasterLink :TMasterDataLink;
    SavedOnMasterChange :TNotifyEvent;
  Begin
    // Some calculated master fields could depend on detail records
    If (Master <> Nil) And Master.Active And MasterEventOnOpen Then
    Begin
      MasterLink := ghMasterLink (Self);
      SavedOnMasterChange := MasterLink.OnMasterChange;

      { We avoid unnecessary call to TCustomClientDataSet.MasterChanged
        method, when master data set processes the deDataSetChange event
        sent from here }
      MasterLink.OnMasterChange := Nil;

      Try
        TDataSetAccess (Master).DataEvent (deDataSetChange, 0);
      Finally
        MasterLink.OnMasterChange := SavedOnMasterChange;
      End;
    End;

    Inherited DoAfterOpen;
  End;

  Procedure TghClientDataSet.DoAfterPost;
  Begin
    FPosting := False;
    Inherited DoAfterPost;
  End;

  // Based on TCustomClientDataSet.DoApplyUpdates method
  Function TghClientDataSet.DoApplyUpdates (Delta :OLEVariant;
    MaxErrors :Integer; Out ErrorCount :Integer) :OLEVariant;
  Var
    OwnerData :OLEVariant;
  Begin
    DoBeforeApplyUpdates (OwnerData);
    Result := AppServer.AS_ApplyUpdates (ProviderName, Delta, MaxErrors,
      ErrorCount, OwnerData);
    FUpdateErrors := ErrorCount;
    DoAfterApplyUpdates (OwnerData);
  End;

  Procedure TghClientDataSet.DoBeforeApplyUpdates (
    Var OwnerData :OLEVariant);
  Begin
    Inherited DoBeforeApplyUpdates (OwnerData);
    CheckIdentify (OwnerData);

    If (GHTransSavePoints <> Nil) And (GHTransSavePoints.Count = 0) Then
      GHTransSavePoints.Save;
  End;

  Procedure TghClientDataSet.DoBeforeDelete;
  Begin
    CheckWaitingForApplyingConfirmation;

    If Not QuietDeleting Then
      Inherited DoBeforeDelete;
  End;

  Procedure TghClientDataSet.DoBeforeEdit;
  Begin
    CheckWaitingForApplyingConfirmation;
    Inherited DoBeforeEdit;
  End;
  
  Procedure TghClientDataSet.DoBeforeExecute (
    Var OwnerData :OLEVariant);
  Begin
    Inherited DoBeforeExecute (OwnerData);
    CheckIdentify (OwnerData);
  End;

  Procedure TghClientDataSet.DoBeforeGetParams (
    Var OwnerData :OLEVariant);
  Begin
    Inherited DoBeforeGetParams (OwnerData);
    CheckIdentify (OwnerData);
  End;

  Procedure TghClientDataSet.DoBeforeGetRecords (
    Var OwnerData :OLEVariant);
  Begin
    Inherited DoBeforeGetRecords (OwnerData);
    CheckIdentify (OwnerData);
  End;

  Procedure TghClientDataSet.DoBeforeInsert;
  Begin
    CheckWaitingForApplyingConfirmation;
    Inherited DoBeforeInsert;
  End;

  Procedure TghClientDataSet.DoBeforeRowRequest (
    Var OwnerData :OLEVariant);
  Begin
    Inherited DoBeforeRowRequest (OwnerData);
    CheckIdentify (OwnerData);
  End;

  Function TghClientDataSet.DoGetRecords (Count :Integer;
    Out RecsOut :Integer; Options :Integer; Const CommandText :WideString;
    Params :OLEVariant) :OLEVariant;
  Var
    AMasterFields :TList;
    I, J :Integer;

    Procedure SetParam (Const Field :String; Const Value :Variant;
      Const AType :TFieldType);
    Var
      Param :String;
    Begin
      // Field name of IndexFieldNames/IndexName to provider parameter name
      Param := ghLeft (GetFieldParamNames (Field), ';');

      If AllParamsWhenDetail Then
        Self.Params.ParamByName (Param).Value := Value
      Else
        Params [I] := VarArrayOf([Param, Value, AType]);
    End;
  Begin
    // If data set is detail, we do an improved fetching
    If MasterActive Then
    Begin
      { If TCustomClientDataSet.OpenCursor is the caller (data set is
        opening) }
      If DSBase = Nil Then
      Begin
        AMasterFields := CreateMasterFieldList;

        Try
          If Not AllParamsWhenDetail Then
            Params := ghVarArr (AMasterFields.Count);

          J := 1;

          For I := 0 To AMasterFields.Count - 1 Do
            With FieldByName (ExtractFieldName (SortFieldNames, J)) Do
              SetParam (FieldName, TField (AMasterFields [I]).Value,
                DataType);
        Finally
          AMasterFields.Free;
        End;

        { PacketRecords = 0 without effect here, we fetch not only
          metadata.  NOTE: In CheckDetailRecords we avoid a second call to
          DoGetRecords (OpenCursor...InternalOpen...CheckDetailRecords). }
        Count := -1;
      End
      Else
        { If TCustomClientDataSet.CheckDetailRecords is the caller (data
          set is already open) }
        If CheckingDetailRecs Then
          For I := 0 To ghFieldCount (MasterFields) - 1 Do
            SetParam (Params [I] [0], Params [I] [1], Params [I] [2]);

      If AllParamsWhenDetail Then
        { This will cause all parameters to be package in the inherited
          method }
        Params := Unassigned;
    End;

    Result := Inherited DoGetRecords (Count, RecsOut, Options, CommandText,
      Params);

    If MasterActive And (DSBase = Nil) Then
      // "ProviderEOF := False" in TCustomClientDataSet.OpenCursor
      RecsOut := PacketRecords;
  End;

  Function TghClientDataSet.GetRecordCount :Integer;
  Begin
    If Active And Not (BOF And EOF) Then
      Result := Inherited GetRecordCount
    Else
      Result := 0;
  End;

  Procedure TghClientDataSet.InternalCancel;
  Begin
    If State = dsInsert Then
      CheckRemoveDetails;

    Inherited InternalCancel;
  End;

  Procedure TghClientDataSet.InternalClose;
  Begin
    Inherited InternalClose;
    ChangeNotificationReg := False;

    // FSavePoints is freed in the destructor
    If (FSavePoints = Nil) Or (csDestroying In ComponentState) Then
      InvalidateNestedSavePoints
    Else
      InvalidateSavePoints;
  End;

  Procedure TghClientDataSet.InternalDelete;
  Var
    Rec :Cardinal;

    Procedure DoInherited;
    Var
      SavedUpdatingCursor :Boolean;
    Begin
      SavedUpdatingCursor := SetUpdatingCursor;

      Try
        Inherited InternalDelete;
      Finally
        UpdatingCursor := SavedUpdatingCursor;
      End;
    End;
  Begin
    InternalDeleting := True;

    Try
      CheckRemoveDetails;

      If ImmediateApplyUpdates Then
      Begin
        DSCursor.GetSequenceNumber (Rec);

        With SavePoints.Save () Do
          Try
            DoInherited;

            If (ApplyChangeLog (ghtrFalse) = 0) Or (ChangeCount = 0) Then
              Confirm
              ///revisar efecto en TDataSet.Delete->Resync que se va al
              //primer registro despus de eliminar aplicando
            Else
              Abort;
          Except
            Restore;
            DSCursor.MoveToSeqNo (Rec);
            Resync([rmCenter]);
            Raise;
          End
      End
      Else
        DoInherited;
    Finally
      InternalDeleting := False;
    End;
  End;

  Procedure TghClientDataSet.InternalPost;
  Var
    Rec :Cardinal;

    Procedure DoInherited;
    Var
      SavedModified, SavedUpdatingCursor :Boolean;
    Begin
      // Don't post existing rows not changed
      If (State = dsEdit) And Not Modified Then
        Exit;

      SavedModified := Modified;
      SavedUpdatingCursor := SetUpdatingCursor;

      Try
        Inherited InternalPost;
      Finally
        UpdatingCursor := SavedUpdatingCursor;

        { Recover Modified property changed by InternalPost ->
          DSCursor.ModifyRecord -> CalcFieldsCallBack -> SetTempState }
        SetModified (SavedModified);
      End;
    End;
  Begin
    If ImmediateApplyUpdates Then
    Begin
      DSCursor.GetSequenceNumber (Rec);

      With SavePoints.Save () Do
        Try
          DoInherited;
          ApplyingInternalPost := True;
          LockResync;  // ApplyChangeLog -> Reconcile -> Resync

          Try
            If (ApplyChangeLog (ghtrFalse) = 0) Or ((ChangeCount = 0) And
            (Not AutoApplyDetails Or (DetailChangeCount () = 0))) Then
              Confirm
            Else
              Abort;
          Finally
            UnLockResync;
            ApplyingInternalPost := False;
          End;
        Except
          Restore;
          DSCursor.MoveToSeqNo (Rec);
          Raise;
        End;
    End
    Else
      DoInherited;
  End;

  Procedure TghClientDataSet.InternalRefresh;
  Begin
    If StatusFilter = [] Then
    Begin
      Inherited InternalRefresh;
      InvalidateSavePoints;
    End
    Else
      { It is necessary to refresh the internal dspropChangeIndex_View
        property (represented by StatusFilter) if an operation was done
        with other cursor that share the same data, because the data set
        can present records incorrectly (possible DataSnap bug). }
      UpdateChangeIndex (StatusFilterAttrs);
  End;

  Procedure TghClientDataSet.Notification (AComponent :TComponent;
    Operation :TOperation);
  Begin
    Inherited Notification (AComponent, Operation);

    If Operation = opRemove Then
      If AComponent = Provider Then
        Provider := Nil
      Else
        If (AComponent Is TDataSet) And (FDetailList <> Nil) Then
          FDetailList.Remove (AComponent);
  End;


  Procedure TghClientDataSet.OpenCursor (InfoQuery :Boolean);
  Begin
    ///revisar por qu
    { NOTA: Al utilizar la propiedad Provider, es necesario reestablecer el
      proveedor con el mtodo SetProvider antes de intentar la apertura. }
    If Provider <> Nil Then
      SetProvider (Provider);

    Inherited OpenCursor (InfoQuery);
  End;

  Procedure TghClientDataSet.SetDataSetField (
    Const Value :TDataSetField);
  Begin
    Inherited SetDataSetField (Value);

    // Mutually exclusive properties
    If Value <> Nil Then
      ImmediateApplyUpdates := False;
  End;

  Procedure TghClientDataSet.SetFieldData (Field :TField;
    Buffer :Pointer);
  Var
    Value :Variant;
  Begin
    If (State = dsBrowse) And AutoEdit Then
      Edit;

    If (State In ghdsEditing) And MustNotifyFieldChange (Field) Then
    Begin
      // New value for BeforeFieldChange
      Value := NativeValue (Field, Buffer);

      If Not CanCheckFieldChange (Field) Or
      Not ghEquals (Field.Value, Value) Then
      Begin
        DoBeforeFieldChange (Field, Value);
        Value := Field.Value;  // Previous value for AfterFieldChange
        Inherited SetFieldData (Field, Buffer);
        DoAfterFieldChange (Field, Value);
        Exit;
      End;
    End;

    Inherited SetFieldData (Field, Buffer);
  End;

  { Public static methods }

  Function TghClientDataSet.ApplyChangeLog (ConfirmSuccess :TghTernary)
    :Integer;
  Begin
    Result := ApplyChangeLog (0, ConfirmSuccess);
  End;

  Function TghClientDataSet.ApplyDetailUpdates (MaxErrors :Integer = 0;
    Const ConfirmSuccess :TghTernary = ghtrDefault) :Integer;
  Var
    I :Integer;
  Begin
    If MaxErrors = -1 Then
      MaxErrors := MaxInt;

    Result := 0;

    For I := 0 To NestedCount - 1 Do
    Begin
      Inc (Result, Nesteds [I].ApplyDetailUpdates (MaxErrors - Result,
        ConfirmSuccess));

      If Result > MaxErrors Then
        Exit;
    End;

    For I := 0 To DetailCount (True) - 1 Do
      If (Details [I] Is TghClientDataSet) And (Details [I].Active) Then
        With TghClientDataSet (Details [I]) Do
        Begin
          Inc (Result, ApplyUpdates (MaxErrors - Result, ConfirmSuccess));

          If Result > MaxErrors Then
            Exit;

          Inc (Result, ApplyDetailUpdates (MaxErrors - Result,
            ConfirmSuccess));

          If Result > MaxErrors Then
            Exit;
        End;
  End;

  Function TghClientDataSet.ApplyUpdates (Const MaxErrors :Integer;
    Const ConfirmSuccess :TghTernary) :Integer;
  Var
    DataSet :TghClientDataSet;
  Begin
    CheckBrowseMode;
    DataSet := Self;

    While DataSet.Master Is TghClientDataSet Do
    Begin
      DataSet := TghClientDataSet (DataSet.Master);

      If DataSet.ApplyingInternalPost And DataSet.AutoApplyDetails Then
      Begin
        // Detail save point of the saved one by master InternalPost
        SavePoints.Save;

        Break;
      End;
    End;

    Result := ApplyChangeLog (MaxErrors, ConfirmSuccess);
  End;

  Function TghClientDataSet.ApplyUpdates (
    Const ConfirmSuccess :TghTernary = ghtrDefault) :Integer;
  Begin
    Result := ApplyUpdates (0, ConfirmSuccess);
  End;

  Procedure TghClientDataSet.CancelDetails;
  Var
    I :Integer;
  Begin
    For I := 0 To NestedCount - 1 Do
    Begin
      Nesteds [I].CancelDetails;
      Nesteds [I].Cancel;
    End;

    For I := 0 To DetailCount (True) - 1 Do
    Begin
      If Details [I] Is TghClientDataSet Then
        TghClientDataSet (Details [I]).CancelDetails;

      Details [I].Cancel;
    End;
  End;

  Procedure TghClientDataSet.CancelDetailUpdates;
  Var
    I :Integer;

    Procedure CancelDataSet (Const DataSet :TghClientDataSet);
    Var
      SavedAutoCancelDetails :Boolean;
    Begin
      SavedAutoCancelDetails := DataSet.AutoCancelDetails;
      DataSet.AutoCancelDetails := True;

      Try
        DataSet.CancelUpdates;
      Finally
        DataSet.AutoCancelDetails := SavedAutoCancelDetails;
      End;
    End;
  Begin
    For I := 0 To NestedCount - 1 Do
      CancelDataSet (Nesteds [I]);

    For I := 0 To DetailCount (True) - 1 Do
      If Details [I] Is TghClientDataSet Then
        CancelDataSet (TghClientDataSet (Details [I]));
  End;

  Function TghClientDataSet.CancelRec :TDataSetState;
  Begin
    Result := State;
    Cancel;
  End;

  Function TghClientDataSet.CloneCursor (Const Reset :Boolean = True;
    Const KeepSettings :Boolean = False) :TghClientDataSet;
  Begin
    Result := TghClientDataSet.Create (Self);
    Result.CloneCursor (Self, Reset, KeepSettings);
  End;

  Function TghClientDataSet.CloneCursorAux (Reset :Boolean = True;
    KeepSettings :Boolean = False) :TghClientDataSet;
  Begin
    Result := CloneCursor (Reset, KeepSettings);
    Result.FIsAuxClone := True;
  End;

  Function TghClientDataSet.ControlsDisabled (
    Const IncludeBlockReadState :Boolean = False) :Boolean;
  Begin
    Result := Inherited ControlsDisabled Or
      (IncludeBlockReadState And (State = dsBlockRead));
  End;

  Function TghClientDataSet.CreateFieldList (Const Names :String)
    :TList;
  Begin
    Result := TList.Create;

    Try
      GetFieldList (Result, Names);
    Except
      Result.Free;
      Raise
    End;
  End;
  
  Function TghClientDataSet.CreateIndexFieldList :TList;
  Begin
    Result := TList.Create;

    Try
      GetIndexFieldList (Result);
    Except
      Result.Free;
      Raise
    End;
  End;

  Function TghClientDataSet.CreateMasterFieldList :TList;
  Begin
    Result := TList.Create;

    Try
      GetMasterFieldList (Result);
    Except
      Result.Free;
      Raise
    End;
  End;

  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.Delete;
  Begin
    FDeleting := True;

    Try
      { We avoid the posting of detail data sets (TDataSet.Delete ->
        DataEvent -> details's CheckBrowseMode method) }
      If AutoCancelDetails Then
        CancelDetails;
        
      Inherited Delete;
    Finally
      FDeleting := False;
    End;
  End;

  Procedure TghClientDataSet.DeleteDetails;
  Var
    I :Integer;

    Procedure DeleteDataSet (Const DataSet :TghClientDataSet);
    Var
      SavedAutoDeleteDetails :Boolean;
    Begin
      SavedAutoDeleteDetails := DataSet.AutoDeleteDetails;
      DataSet.AutoDeleteDetails := True;

      Try
        DataSet.Cancel;

        If DataSet.MasterActive Then  // Detail data set
          DataSet.DeleteRecsDetail (Self, Nil, Null)
        Else  // Nested data set
          DataSet.DeleteRecsNested (RecNumber, Nil, Null);
      Finally
        DataSet.AutoDeleteDetails := SavedAutoDeleteDetails;
      End;
    End;
  Begin
    For I := 0 To Primary.NestedCount - 1 Do
      DeleteDataSet (Primary.Nesteds [I]);

    For I := 0 To Primary.DetailCount (True) - 1 Do
      If (Primary.Details [I] Is TghClientDataSet) And
      (Primary.Details [I].Active) Then
        DeleteDataSet (TghClientDataSet (Primary.Details [I]));
  End;

  Procedure TghClientDataSet.DeleteQuiet;
  Begin
    Inc (QuietDeletingCount);

    Try
      Delete;
    Finally
      Dec (QuietDeletingCount);
    End;
  End;

  Procedure TghClientDataSet.DeleteRecs (
    Const VisibleOnly :Boolean = False;
    Const FilterProc :TghEnumProc = Nil);
  Begin
    DeleteRecs (VisibleOnly, FilterProc, Null);
  End;

  Procedure TghClientDataSet.DeleteRecs (Const FilterProc :TghEnumProc;
    Const FilterParams :Variant);
  Begin
    DeleteRecs (False, FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.DeleteRecs (Const FilterProc :TghEnumProc);
  Begin
    DeleteRecs (False, FilterProc, Null);
  End;

  Function TghClientDataSet.DetailChangeCount :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To NestedCount - 1 Do
      Inc (Result, Nesteds [I].DetailChangeCount ());

    For I := 0 To DetailCount (True) - 1 Do
      If Details [I] Is TghClientDataSet Then
        With TghClientDataSet (Details [I]) Do
          Inc (Result, ChangeCount + DetailChangeCount ());
  End;

  Function TghClientDataSet.CloneAsDetail (
    Const AMasterSource :TComponent; Const AMasterFields :String;
    Const AIndexFieldNames :String = '';
    Const AFieldParamNames :String = '') :TghClientDataSet;
  Begin
    Result := TghClientDataSet (ghCloneDataSet (Self, Self,
      'Active;MasterSource;MasterFields;IndexFieldNames;FieldParamNames'));

    If AMasterSource Is TDataSource Then
      Result.MasterSource := TDataSource (AMasterSource)
    Else
      If AMasterSource Is TDataSet Then
      Begin
        Result.MasterSource := LinkDataSourceClass.Create (Result);
        Result.MasterSource.DataSet := TDataSet (AMasterSource);
      End;

    Result.MasterFields := AMasterFields;

    If AIndexFieldNames <> '' Then
      Result.IndexFieldNames := AIndexFieldNames
    Else
      Result.IndexFieldNames := AMasterFields;

    Result.FieldParamNames := AFieldParamNames
  End;

  Function TghClientDataSet.DetailCount (Const Update :Boolean = False)
    :Integer;
  Begin
    Result := DetailList (Update).Count;
  End;

  Function TghClientDataSet.DisableControls (
    Const IncludeNonvisual :Boolean = True) :TBookmark;
  Begin
    Inherited DisableControls;

    If BlockReadSizeStack = Nil Then
      BlockReadSizeStack := TStack.Create;

    // We save current BlockReadSize to restore it in EnableControls
    BlockReadSizeStack.Push (Pointer (BlockReadSize));

    { A value of 1 puts the data set in the dsBlockRead state (visual
      controls are not notified, but detail links are) }
    BlockReadSize := Byte (Not IncludeNonvisual);

    Result := GetBookmark;
  End;

  Procedure TghClientDataSet.Empty (
    Const AProviderEOF :TghTernary = ghtrDefault);
  Begin
    If Active Then
    Begin
      EmptyDataSet;
      InvalidateSavePoints;
      ProviderEOF := (AProviderEOF = ghtrTrue) Or
        ((AProviderEOF = ghtrDefault) And Not MasterActive);
    End;

    If AutoCancelDetails Then
      EmptyDetails (AProviderEOF);
  End;

  Procedure TghClientDataSet.EmptyDetails (
    Const AProviderEOF :TghTernary = ghtrDefault);
  Var
    I :Integer;

    Procedure DoEmpty (Const DataSet :TghClientDataSet);
    Var
      SavedAutoCancelDetails :Boolean;
    Begin
      SavedAutoCancelDetails := DataSet.AutoCancelDetails;
      DataSet.AutoCancelDetails := True;

      Try
        DataSet.Empty (AProviderEOF);
      Finally
        DataSet.AutoCancelDetails := SavedAutoCancelDetails;
      End;
    End;
  Begin
    For I := 0 To NestedCount - 1 Do
      DoEmpty (Nesteds [I]);

    For I := 0 To DetailCount (True) - 1 Do
      If Details [I] Is TghClientDataSet Then
        DoEmpty (TghClientDataSet (Details [I]));
  End;

  Procedure TghClientDataSet.EnableControls (
    Const ABookmark :TBookmark = Nil);
  Begin
    GoToBookmarkValid (ABookmark);

    If (BlockReadSizeStack <> Nil) And (BlockReadSizeStack.Count > 0) Then
      BlockReadSize := Integer (BlockReadSizeStack.Pop);

    Inherited EnableControls;
  End;

  Function TghClientDataSet.EnumDataSources (Const Routine :TghEnumProc;
    Const AParams :Variant) :TghDataSource;

    // Call back routine
    Procedure CheckLinked (Const DataSource :TghDataSource;
      Const AParams :Variant; Var AContinue :Boolean);
    Begin
      AContinue := (DataSource.DataSet <> ghPtr (AParams [2])) Or
        ghCallNestedEnumProc (AParams, DataSource);
    End;
  Begin
    If GHDataSources <> Nil Then
      Result := GHDataSources.Enum (@CheckLinked,
        ghVar ([@Routine, @AParams, Self])) As TghDataSource
    Else
      Result := Nil;
  End;

  Function TghClientDataSet.GetLogChangesQuiet :Boolean;
  Begin
    CheckActive;
    Result := Boolean (BaseProp (dspropLogChanges));
  End;

  Function TghClientDataSet.GoToBookmarkValid (Const Value :TBookmark)
    :Boolean;
  Begin
    Result := BookmarkValid (Value);

    If Result Then
      GoToBookmark (Value);
  End;
  
  Function TghClientDataSet.HasMasterSavePoint :Boolean;
  Var
    I :Integer;
  Begin
    If (ghSavePoints <> Nil) And Not HasSavePoint (ghSavePoints [0]) Then
      For I := 1 To ghSavePoints.Count - 1 Do
        If HasSavePoint (ghSavePoints [I]) Then
        Begin
          Result := True;
          Exit;
        End;

    Result := False;
  End;

    Type
      TDataLinkAccess = Class (TDataLink);
  Function TghClientDataSet.HasVisualLinks :Boolean;

    // Call back routine
    Procedure CheckDataSource (Const DataSource :TghDataSource;
      Const Unused :Variant; Var AContinue :Boolean);
    Var
      I :Integer;
    Begin
      For I := 0 To DataSource.DataLinks.Count - 1 Do
        If TDataLinkAccess (DataSource.DataLinks [I]).VisualControl Then
        Begin
          AContinue := False;
          Exit;
        End;
    End;
  Begin
    Result := EnumDataSources (@CheckDataSource, Null) <> Nil;
  End;

  Function TghClientDataSet.InvalidateSavePoints (
    Const StartGlobalIndex :Integer = 0;
    Const DetailAction :TghSavePointAction = ghsaNone) :Integer;
  Var
    ASavePoints :TghClientDataSetSavePoints;
  Begin
    ASavePoints := FindSavePoints;

    If ASavePoints <> Nil Then
      Result := ASavePoints.InvalidateGlobal (StartGlobalIndex)
    Else
      Result := -1;

    Result := ghMinCardinal (Result,
      InvalidateNestedSavePoints (StartGlobalIndex));

    If (Result > -1) And (DetailAction <> ghsaNone) Then
      ghSavePoints.Terminate (Result, DetailAction = ghsaConfirm);
  End;

  Function TghClientDataSet.LinkDataSource :TghDataSource;
  Begin
    Result := EnumDataSources (Nil, Null);  // First associated data source

    If Result = Nil Then
    Begin
      Result := LinkDataSourceClass.Create (Self);
      Result.DataSet := Self;
    End;
  End;
  
  Function TghClientDataSet.Locate (Const KeyFields :String;
    Const KeyValues :Variant) :Boolean;
  Begin
    Result := Locate (KeyFields, KeyValues, []);
  End;

  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.MergeChangeLog;
  Begin
    InternalMergeChangeLog;
    InvalidateSavePoints;
    WaitingForApplyingConfirmation := False;
  End;

  Procedure TghClientDataSet.MergeDetailChanges;
  Var
    I :Integer;

    Procedure MergeDataSet (Const DataSet :TghClientDataSet);
    Var
      SavedAutoApplyDetails :Boolean;
    Begin
      SavedAutoApplyDetails := DataSet.AutoApplyDetails;
      DataSet.AutoApplyDetails := True;

      Try
        DataSet.MergeChangeLog;
      Finally
        DataSet.AutoApplyDetails := SavedAutoApplyDetails;
      End;
    End;
  Begin
    For I := 0 To NestedCount - 1 Do
      MergeDataSet (Nesteds [I]);

    For I := 0 To DetailCount (True) - 1 Do
      If (Details [I] Is TghClientDataSet) And (Details [I].Active) Then
        MergeDataSet (TghClientDataSet (Details [I]));
  End;

  Procedure TghClientDataSet.PostDetails;
  Var
    I :Integer;

    Procedure PostDataSet (Const DataSet :TDataSet);
    Begin
      If DataSet Is TghClientDataSet Then
        TghClientDataSet (DataSet).PostDetails;
        
      If DataSet.State In dsEditModes Then
        DataSet.Post;
    End;
  Begin
    For I := 0 To NestedCount - 1 Do
      PostDataSet (Nesteds [I]);

    For I := 0 To DetailCount (True) - 1 Do
      PostDataSet (Details [I]);
  End;

  Function TghClientDataSet.QuietDeleting :Boolean;
  Begin
    Result := (QuietDeletingCount > 0) Or ((Parent <> Nil) And
      Parent.QuietDeleting) Or ((Master Is TghClientDataSet) And
      TghClientDataSet (Master).QuietDeleting);
  End;

  Function TghClientDataSet.RecAttrs :Integer;
  Var
    ABookmark :TBookMark;
    SavedStatusFilter :TUpdateStatusSet;
  Begin
    If State = dsInsert Then
      Result := dsRecNew
    Else
    Begin
      UpdateCursorPos;
      Result := CursorRecAttrs (DSCursor);

      { For new records, DSCursor.GetRecordAttribute returns the dsRecNew
        attribute only, when the record has been modified or modified and
        then deleted (DataSnap bug).  This code adds the missing
        attributes. }
      If Result = dsRecNew Then
      Begin
        ABookmark := GetBookmark;
        SavedStatusFilter := ChangeLog.SetTempStatusFilter ([usModified]);

        Try
          If ChangeLog.BookmarkValid (ABookMark) Then
          Begin
            Result := Result Or dsRecModified;
            ChangeLog.StatusFilter := [usDeleted];

            If ChangeLog.BookmarkValid (ABookMark) Then
              Result := Result Or dsRecDeleted;
          End;
        Finally
          ChangeLog.RestoreStatusFilter (SavedStatusFilter);

          If ChangeLog = Self Then
            GoToBookmark (ABookmark);
        End;
      End;
    End;
  End;

  Procedure TghClientDataSet.RestoreFilter (
    Const Filtering :TghDataSetFiltering);
  Begin
    Try
      Filtered := False;
      Filter := Filtering.Expr;
      OnFilterRecord := Filtering.Event;
      Filtered := Filtering.Active;
    Finally
      EnableControls (Filtering.Bookmark);
    End;
  End;

  Procedure TghClientDataSet.RestoreStatusFilter (
    Const Value :TUpdateStatusSet);
  Begin
    Try
      StatusFilter := Value;
    Finally
      EnableControls;
    End;
  End;

  Procedure TghClientDataSet.RevertChanges (
    AUpdateStatus :TghChangeUpdateStatusSet;
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Var
    SavedAutoCancelDetails, SavedAutoDeleteDetails :Boolean;
    SavedStatusFilter :TUpdateStatusSet Absolute AUpdateStatus;
  Begin
    If AUpdateStatus = [] Then
      AUpdateStatus := ghusChanges;

    SavedStatusFilter := ChangeLog.SetTempStatusFilter (AUpdateStatus);
    SavedAutoCancelDetails := ChangeLog.AutoCancelDetails;
    ChangeLog.AutoCancelDetails := AutoCancelDetails;
    SavedAutoDeleteDetails := ChangeLog.AutoDeleteDetails;
    ChangeLog.AutoDeleteDetails := AutoDeleteDetails;

    Try
      ChangeLog.RevertRecsVisible (FilterProc, FilterParams);
    Finally
      ChangeLog.AutoDeleteDetails := SavedAutoDeleteDetails;
      ChangeLog.AutoCancelDetails := SavedAutoCancelDetails;
      ChangeLog.RestoreStatusFilter (SavedStatusFilter);
    End;
  End;

  Procedure TghClientDataSet.RevertChanges (
    Const AUpdateStatus :TghChangeUpdateStatusSet = ghusChanges;
    Const FilterProc :TghEnumProc = Nil);
  Begin
    RevertChanges (AUpdateStatus, FilterProc, Null);
  End;

  Procedure TghClientDataSet.RevertChanges (
    Const FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    RevertChanges (ghusChanges, FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.RevertChanges (
    Const FilterProc :TghEnumProc);
  Begin
    RevertChanges (ghusChanges, FilterProc, Null);
  End;

  Procedure TghClientDataSet.RevertDetails;
  Var
    I :Integer;

    Procedure RevertDataSet (Const DataSet :TghClientDataSet);
    Var
      SavedAutoCancelDetails :Boolean;
    Begin
      SavedAutoCancelDetails := DataSet.AutoCancelDetails;
      DataSet.AutoCancelDetails := True;

      Try
        DataSet.Cancel;
        
        If DataSet.MasterActive Then  // Detail data set
          DataSet.RevertRecsDetail (Self, Nil, Null)
        Else  // Nested data set
          DataSet.RevertRecsNested (RecNumber, Nil, Null);
      Finally
        DataSet.AutoCancelDetails := SavedAutoCancelDetails;
      End;
    End;
  Begin
    For I := 0 To Primary.NestedCount - 1 Do
      RevertDataSet (Primary.Nesteds [I]);

    For I := 0 To Primary.DetailCount (True) - 1 Do
      If (Primary.Details [I] Is TghClientDataSet) And
      (Primary.Details [I].Active) Then
        RevertDataSet (TghClientDataSet (Primary.Details [I]));
  End;

  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.RevertRecord;
  Var
    Attrs :Byte;
    AuxCursor :IDSCursor;
    SavedUpdatingCursor :Boolean Absolute Attrs;
  Begin
    If CancelRec = dsInsert Then
      Exit;

    CheckWaitingForApplyingConfirmation;
    CheckBrowseMode;
    CheckRevertDetails;
    Attrs := RecAttrs;  // RecAttrs calls the UpdateCursorPos method

    If Not ghInBits (Attrs, ghdaRecAttrsChanges) Then
      Exit;

    If ghInBits (Attrs, dsRecNew Or dsRecModified) Then
    Begin
      If ghHasBit (Attrs, dsRecNew) Then
        CheckDeleteDetails;

      AuxCursor := CreateCursor (GetBookmark);

      { NOTE: IDSCursor.RevertRecord don't work properly with new records
        in filtered cursors.  The record reappears as "unmodified" when the
        Filtered property changes, except if the record is the last logged
        operation (DataSnap bug).  In addition, the modified records with
        changes in internally calculated fields only, are not affect the
        change count (but the change log are) and they are not reversible
        by IDSCursor.RevertRecord (possible and inadequate DataSnap
        convention). }
      If ChangeCurrentRec (AuxCursor, ghHasBit (Attrs, dsRecNew)) Then
      Begin
        Check (AuxCursor.RevertRecord);
        InvalidateSavePoints;

        { NOTE: IDSCursor.RevertRecord calls (indirectly) the
          InternalChangeNotification method; even so, we must call the
          RefreshChangeLogs method here in order to refresh the ChangeLog
          data sets (perhaps due to a temporary lock inside of
          IDSCursor.RevertRecord). }
        RefreshChangeLogs;

        Exit;
      End;
    End;

    SavedUpdatingCursor := SetUpdatingCursor;

    Try
      Check (DSCursor.RevertRecord);
    Finally
      UpdatingCursor := SavedUpdatingCursor;
    End;

    InvalidateSavePoints;
    Resync ([]);
    RefreshChangeLogs (False);  // See note above
  End;

  Procedure TghClientDataSet.RevertRecs (
    Const VisibleOnly :Boolean = False;
    Const FilterProc :TghEnumProc = Nil);
  Begin
    RevertRecs (VisibleOnly, FilterProc, Null);
  End;

  Procedure TghClientDataSet.RevertRecs (Const FilterProc :TghEnumProc;
    Const FilterParams :Variant);
  Begin
    RevertRecs (False, FilterProc, FilterParams);
  End;

  Procedure TghClientDataSet.RevertRecs (Const FilterProc :TghEnumProc);
  Begin
    RevertRecs (False, FilterProc, Null);
  End;

  Function TghClientDataSet.RowCount (
    Const VisibleOnly :Boolean = False;
    Const FilterProc :TghEnumProc = Nil) :Integer;
  Begin
    Result := RowCount (VisibleOnly, FilterProc, Null);
  End;

  Function TghClientDataSet.RowCount (Const FilterProc :TghEnumProc;
    Const FilterParams :Variant) :Integer;
  Begin
    Result := RowCount (False, FilterProc, FilterParams);
  End;

  Function TghClientDataSet.RowCount (Const FilterProc :TghEnumProc)
    :Integer;
  Begin
    Result := RowCount (False, FilterProc, Null);
  End;

  Function TghClientDataSet.SetLogChangesQuiet (
    Const Value :Boolean = True) :Boolean;
  Begin
    Result := GetLogChangesQuiet;

    If Value = Result Then
      Exit;

    Check (DSBase.SetProp (dspropLogChanges, Integer (Value)));

    If Value Then
      { It is necessary to rebuild the szChangeIndex index.  This avoids
        the "operation not applicable" error of a subsequent call to the
        IDSCursor.ModifyRecord method (which happens with records that
        already have modifications in the change log). }
      UpdateChangeIndex;
  End;

  Function TghClientDataSet.SetTempFilter (Const Expr :String;
    Const LockNonvisualLinks :Boolean = True) :TghDataSetFiltering;
  Begin
    Result := ghFiltering (Self);
    DisableControls (LockNonvisualLinks);

    Try
      Filtered := False;
      Filter := Expr;
      OnFilterRecord := Nil;
      Filtered := Expr <> '';
    Except
      RestoreFilter (Result);
      Raise;
    End;
  End;

  Function TghClientDataSet.SetTempStatusFilter (
    Const Value :TUpdateStatusSet;
    Const LockNonvisualLinks :Boolean = True) :TUpdateStatusSet;
  Begin
    Result := StatusFilter;
    DisableControls (LockNonvisualLinks);

    Try
      StatusFilter := Value;
    Except
      RestoreStatusFilter (Result);
      Raise;
    End;
  End;

  { Public virtual methods }

  Function TghClientDataSet.ApplyChangeLog (MaxErrors :Integer = 0;
    ConfirmSuccess :TghTernary = ghtrDefault) :Integer;
  Var
    Rows :OLEVariant;
    SavedCount :Integer;
    SavedModified :Boolean;
    SavedState :TDataSetState;

    Procedure CheckDetails;
    Begin
      If AutoApplyDetails And Not InternalDeleting Then
        Result := RootParent.ApplyDetailUpdates (MaxErrors,
          ConfirmSuccess);
    End;
  Begin
    With RootParent Do
    Begin
      Result := 0;

      If ChangeCount > 0 Then
      Begin
        If Not WaitingForApplyingConfirmation Then
        Begin
          { SetTempState and RestoreState sets Modified to False (we could
            be in InternalPost) }
          SavedModified := Modified;

          { Avoid recursion GetDelta -> CheckBrowseMode -> Post ->
            InternalPost -> ApplyChangeLog }
          SavedState := SetTempState (dsBrowse);

          Try
            Rows := Delta;
          Finally
            RestoreState (SavedState);
            SetModified (SavedModified);
          End;

          Rows := DoApplyUpdates (Rows, MaxErrors, Result);
        End;

        If Result = 0 Then
        Begin
          CheckDetails;

          If (ConfirmSuccess = ghtrTrue) Or ((ConfirmSuccess =
          ghtrDefault) And Not HasMasterSavePoint ()) Then
            MergeChangeLog
          Else
            WaitingForApplyingConfirmation := True;
        End
        Else
        Begin
          SavedCount := ChangeCount;

          { Reconcile changes Modified to False (we could be in
            InternalPost) }
          SavedModified := Modified;

          Try
            Reconcile (Rows);
          Finally
            SetModified (SavedModified);
          End;

          { If some record was applied / reconciled }
          If ChangeCount < SavedCount Then
            InvalidateSavePoints;  { Invalidate all save points }
        End;
      End
      Else
        CheckDetails;
    End;
  End;

  // Suggestion to the VCL writers: make the native method virtual.
  Procedure TghClientDataSet.CancelUpdates;
  Var
    SavedUpdatingCursor :Boolean;
  Begin
    FCanceling := True;

    Try
      If AutoCancelDetails Then
        CancelDetailUpdates;

      If ChangeCount = 0 Then
      Begin
        Cancel;
        FCanceling := True;
      End
      Else
      Begin
        SavedUpdatingCursor := SetUpdatingCursor;

        Try
          Inherited CancelUpdates;
        Finally
          UpdatingCursor := SavedUpdatingCursor;
        End;
      End;

      InvalidateSavePoints;
      WaitingForApplyingConfirmation := False;
    Finally
      FCanceling := False;
    End;
  End;

  Procedure TghClientDataSet.DeleteRecs (VisibleOnly :Boolean;
    FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    Cancel;

    If VisibleOnly Then
      DeleteRecsVisible (FilterProc, FilterParams)
    Else
      If DataSetField <> Nil Then
        DeleteRecsNested (Parent.RecNumber, FilterProc, FilterParams){}//probar
      Else
        If MasterActive Then
          DeleteRecsDetail (Master, FilterProc, FilterParams)
        Else
          DeleteRecsFilter (FilterProc, FilterParams)
  End;

  Function TghClientDataSet.GetFieldParamNames (Const Field :String)
    :String;
  Begin
    Result := ghReplace (ghParam (FieldParamNames, Field), ',', ';');

    If (Result = '') And (Params.FindParam (Field) <> Nil) Then
      Result := Field;
  End;

  Function TghClientDataSet.FocusField (Const Field :String) :Boolean;
  Var
    AField :TField;

    // Call back routine
    Procedure CheckDataLinks (Const DataSource :TghDataSource;
      Const Field :Variant; Var AContinue :Boolean);
    Var
      I :Integer;
    Begin
      For I := 0 To DataSource.DataLinks.Count - 1 Do
        If TDataLinkAccess (DataSource.DataLinks [I]).VisualControl Then
        Begin
          TDataLinkAccess (DataSource.DataLinks [I]).DataEvent (
            deFocusControl, Cardinal (Field));

          // If field was focused
          If TFieldRef (Cardinal (Field))^ = Nil Then
          Begin
            AContinue := False;
            Exit;
          End;
        End;
    End;
  Begin
    AField := FieldByName (Field);

    If ControlsDisabled Then
      { Direct deFocusControl message to the visual data links of
        associated TghDataSource objects }
      FieldFocused := ghBeforeContinue (@BeforeFocusField, AField) And
        (EnumDataSources (@CheckDataLinks, ghVar (@AField)) <> Nil)
    Else
      // Normal call to the field's FocusControl method
      ghBeforeContinue (@BeforeFocusField, AField, AField.FocusControl);

    Result := FieldFocused;
  End;

  Procedure TghClientDataSet.RevertRecs (VisibleOnly :Boolean;
    FilterProc :TghEnumProc; Const FilterParams :Variant);
  Begin
    // Selective reversion
    If VisibleOnly Or Assigned (FilterProc) Or Detailed Then
    Begin
      Cancel;

      If VisibleOnly Then
        RevertRecsVisible (FilterProc, FilterParams)
      Else
        If DataSetField <> Nil Then
          RevertRecsNested (Parent.RecNumber, FilterProc, FilterParams){}//probar
        Else
          If MasterActive Then
            RevertRecsDetail (Master, FilterProc, FilterParams)
          Else
            RevertRecsFilter (FilterProc, FilterParams);
    End
    Else
      // Not selective reversion (we revert the entire change log)
      CancelUpdates;
  End;

  Function TghClientDataSet.RowCount (VisibleOnly :Boolean;
    FilterProc :TghEnumProc; Const FilterParams :Variant) :Integer;
  Begin
    CheckBrowseMode;

    // Not selective count
    If Not Assigned (FilterProc) And (VisibleOnly Or Not Limited Or
    (MasterActive And Not Filtered)) Then
      Result := RecordCount
    Else  // Selective count
      If VisibleOnly Then
        Result := RowCountVisible (FilterProc, FilterParams)
      Else
        If DataSetField <> Nil Then
          Result := RowCountNested (Parent.RecNumber, FilterProc,
            FilterParams){}//probar
        Else
          If MasterActive Then
            Result := RowCountDetail (Master, FilterProc, FilterParams)
          Else
            Result := RowCountFilter (FilterProc, FilterParams);
  End;

  { Public overridden methods }

  Function TghClientDataSet.ApplyUpdates (MaxErrors :Integer) :Integer;
  Begin
    Result := ApplyUpdates (MaxErrors, ghtrDefault);
  End;

  Function TghClientDataSet.BookmarkValid (Bookmark :TBookmark) :Boolean;
  Begin
    // Additional necessary validation
    Result := (Bookmark <> Nil) And Inherited BookmarkValid (Bookmark);
  End;

  Procedure TghClientDataSet.Cancel;
  Begin
    FCanceling := True;

    Try
      { We avoid the posting of detail data sets (TDataSet.Cancel ->
        DataEvent -> details's CheckBrowseMode method) }
      If AutoCancelDetails Then
        CancelDetails;

      Inherited Cancel;
    Finally
      FCanceling := False;
    End;
  End;

  Procedure TghClientDataSet.CloneCursor (Source :TCustomClientDataSet;
    Reset :Boolean; KeepSettings :Boolean = False);
  Begin
    { We include FetchOnDemand in properties to clear.  With this, we avoid
      the "Missing data provider or data packet" exception if the source
      cursor is empty and belongs to a detail data set. }
    If Reset Then
      FetchOnDemand := False;

    Inherited CloneCursor (Source, Reset, KeepSettings);

    { The inherited CloneCursor method installs the native notifcation
      call back function }
    ChangeNotificationReg := False;

    RegChangeNotification;

    If Source Is TghClientDataSet Then
    Begin
      TghClientDataSet (Source).ChangeNotificationReg := False;
      TghClientDataSet (Source).RegChangeNotification;
    End;

    // Additional reset actions (completely "clean" cursor)
    If Reset Then
    Begin
      CancelRange;
      Check (DSCursor.UseIndexOrder (szDefault_Order));
    End;
  End;

  Function TghClientDataSet.DataRequest (Data :OLEVariant) :OLEVariant;
  Begin
    CheckIdentify (Data);
    Inherited DataRequest (Data);
  End;

  Procedure TghClientDataSet.GetDetailLinkFields (
    AMasterFields, DetailFields :TList);
  Begin
    GetMasterFieldList (AMasterFields);
    GetIndexFieldList (DetailFields);

    If DetailFields.Count > AMasterFields.Count Then
      DetailFields.Count := AMasterFields.Count;
  End;

    Type
      TFieldAccess = Class (TField);
  Function TghClientDataSet.GetFieldData (Field :TField;
    Buffer :Pointer) :Boolean;
  Begin
    If TempFieldData <> Nil Then
    Begin
      Result := TempFieldData^ <> Nil;

      If Result And (Buffer <> Nil) Then
        TFieldAccess (Field).CopyData (TempFieldData^, Buffer);
    End
    Else
      { When State is dsOldValue and the active record is new (not posted
        yet), the inherited GetFieldData method returns incorrect values
        (bug in the TCustomClientDataSet.GetActiveRecBuf method).  This
        code corrects the problem. }
      If (State = dsOldValue) And
      (GetBookmarkFlag (ActiveBuffer) <> bfCurrent) Then
        // The record does not have a previous/posted version
        Result := False
      Else
        Result := Inherited GetFieldData (Field, Buffer);
  End;

  Function TghClientDataSet.Locate (Const KeyFields :String;
    Const KeyValues :Variant; Options :TLocateOptions): Boolean;
  Var
    Field :String;
    FieldsLength, I :Integer;
  Begin
    If (ghIndex (KeyFields, ';') = 0) Or VarIsArray (KeyValues) Then
      Result := Inherited Locate (KeyFields, KeyValues, Options)
    Else  // Search of a value in different fields
    Begin
      I := 1;
      FieldsLength := Length (KeyFields);

      While I <= FieldsLength Do
      Begin
        Field := ExtractFieldName (KeyFields, I);

        If (Not (FieldByName (Field).DataType In ghftNumerics) Or
        ghIsFloat (KeyValues)) And Inherited Locate (Field, KeyValues,
        Options) Then
        Begin
          Result := True;
          Exit;
        End;
      End;

      Result := False;
    End;
  End;

  Procedure TghClientDataSet.Post;
  Begin
    FPosting := True;

    Try
      Inherited Post;
    Finally
      FPosting := False;
    End;
  End;

  Procedure TghClientDataSet.Resync (Mode :TResyncMode);
  Begin
    { NOTE: Use this condition in the Resync methods overridden in derived
      classes too. }
    If Not ResyncLocked Then
      Inherited Resync (Mode);
  End;

End.

