{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFFileVersionInfo.pas - TghFileVersionInfo class unit.                 }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFFileVersionInfo;  { File Version Information }

Interface

  Uses
    Windows, GHFRTL;

  Type
    { File Version Information class }
    TghFileVersionInfo = Class
      Protected
        { Fields }
        Data :Pointer;
        FPath :String;
        FSize :Integer;

        { Static methods }
        Function GetFixed :PVSFixedFileInfo;
        Function GetTranslationCode (Const Index :Integer) :Integer;
        Function GetTranslationCount :Integer;
        Function GetTranslation (Const Index :Integer) :TghTwoWords;
        Function GetTranslations (Var Ptr :Pointer; Var Count :Integer)
          :Boolean;
        Function GetVersion :Int64;
        Function GetVersionStr :String;
        Function QueryData (Const Key :String; Var Ptr :Pointer;
          ASize :PCardinal = Nil) :Boolean;

        { Virtual methods }
        Function InternalQueryData (Const Key :String; Var Ptr :Pointer;
          Var Size :Cardinal) :Boolean; Virtual;
        Function InternalStr (TranslationCode :Integer; Const Name :String)
          :PChar; Virtual;
      Public
        Constructor Create (Const APath :String); Overload; Virtual;
        Constructor Create (Const Module :THandle); Overload; Virtual;
        Destructor Destroy; Override;

        { Static methods }
        Function OrigFileName :String;
        Function OrigName :String;
        Function OrigPath :String; Overload;
        Function OrigPath (Const Ext :String) :String; Overload;
        Function Str (Const TranslationCode :Integer; Const Name :String)
          :PChar; Overload;
        Function Str (Const Name :String;
          Const TranslationIndex :Integer = 0) :PChar; Overload;

        { Properties }
        Property Fixed :PVSFixedFileInfo Read GetFixed;
        Property Path :String Read FPath;
        Property Size :Integer Read FSize;
        Property TranslationCodes [Const Index :Integer] :Integer
          Read GetTranslationCode;
        Property TranslationCount :Integer Read GetTranslationCount;
        Property Translations [Const Index :Integer] :TghTwoWords
          Read GetTranslation;
        Property Version :Int64 Read GetVersion;
        Property VersionStr :String Read GetVersionStr;
    End;

Implementation

  Uses
    SysUtils;

  { TghFileVersionInfo }

  Constructor TghFileVersionInfo.Create (Const APath :String);
  Begin
    FPath := ghSearchPath (APath, True);

    If ghTheExe (FPath) Then  // Main executable of the current process
      Create (MainInstance)
    Else
      If ghTheModule (FPath) Then  // HInstance module
        Create (HInstance)
      Else
      Begin
        FSize := GetFileVersionInfoSize (PChar (FPath), GHDUCardinal);

        If FSize = 0 Then
          Exit;

        GetMem (Data, FSize);

        If Not GetFileVersionInfo (PChar (FPath), 0, FSize, Data) Then
          ghRaiseLastError;
      End;
  End;

  Constructor TghFileVersionInfo.Create (Const Module :THandle);
  Var
    ASize :Integer;
  Begin
    If FPath = '' Then
      FPath := ghModulePath (Module);

    Data := ghLockResource (Module, vs_Version_Info, rt_Version, @ASize);

    If Data <> Nil Then
    Begin
      // Work size = (SizeOfResource * 2) + 4 = GetFileVersionInfoSize
      FSize := (ASize * 2) + 4;
      Data := ghCopyMem (Data, ASize, FSize);
    End
    Else
      FSize := 0;
  End;

  Destructor TghFileVersionInfo.Destroy;
  Begin
    If Data <> Nil Then
      FreeMem (Data);
  End;

  { Protected static methods }

  Function TghFileVersionInfo.GetFixed :PVSFixedFileInfo;
  Begin
    If Not QueryData ('\', Pointer (Result)) Then
      Result := Nil;
  End;

  Function TghFileVersionInfo.GetTranslationCode (Const Index :Integer)
    :Integer;
  Begin
    Result := ghSwapWords (Translations [Index].IntValue);
  End;

  Function TghFileVersionInfo.GetTranslationCount :Integer;
  Begin
    If Not GetTranslations (GHDUPtr, Result) Then
      Result := 0;
  End;

  Function TghFileVersionInfo.GetTranslation (Const Index :Integer)
    :TghTwoWords;
  Var
    Count :Integer;
    Ptr :Pointer;
  Begin
    If GetTranslations (Ptr, Count) Then
      Result := TghTwoWordsArr (Ptr) [ghCheckIndex (Index, Count)]
    Else
      ghRaiseIndex (Index);
  End;

  Function TghFileVersionInfo.GetTranslations (Var Ptr :Pointer;
    Var Count :Integer) :Boolean;
  Begin
    Result := QueryData ('\VarFileInfo\Translation', Ptr, @Count);

    If Result Then
      Count := Count Div SizeOf (TghTwoWords);
  End;

  Function TghFileVersionInfo.GetVersion :Int64;
  Begin
    If Fixed <> Nil Then
    Begin
      TghTwoInts (Result).HighValue := Fixed.dwFileVersionMS;
      TghTwoInts (Result).LowValue := Fixed.dwFileVersionLS;
    End
    Else
      Result := 0;
  End;

  Function TghFileVersionInfo.GetVersionStr :String;
  Begin
    If Fixed <> Nil Then
      With PghFourWords (@Fixed.dwFileVersionMS)^ Do
        Result := ghConcatInts ([Value2, Value1, Value4, Value3], '.')
    Else
      Result := '';
  End;

  Function TghFileVersionInfo.QueryData (Const Key :String;
    Var Ptr :Pointer; ASize :PCardinal = Nil) :Boolean;
  Begin
    If Data <> Nil Then
    Begin
      If ASize = Nil Then
        ASize := @GHDUCardinal;

      Result := InternalQueryData (Key, Ptr, ASize^);
    End
    Else
      Result := False;
  End;

  { Protected virtual methods }

  Function TghFileVersionInfo.InternalQueryData (Const Key :String;
    Var Ptr :Pointer; Var Size :Cardinal) :Boolean;
  Begin
    Result := VerQueryValue (Data, PChar (Key), Ptr, Size);
  End;

  Function TghFileVersionInfo.InternalStr (TranslationCode :Integer;
    Const Name :String) :PChar;
  Begin
    If Not InternalQueryData (Format ('\StringFileInfo\%.8x\%s',
    [TranslationCode, Name]), Pointer (Result), GHDUCardinal) Then
      Result := Nil;
  End;

  { Public static methods }

  Function TghFileVersionInfo.OrigFileName :String;
  Begin
    // Original file name (OriginalFilename string) or file name in Path

    Result := Str (ghvsOriginalFilename);

    If Result = '' Then
      Result := ExtractFileName (Path);
  End;

  Function TghFileVersionInfo.OrigName :String;
  Begin
    // Original file name without extension
    Result := ghExtPath (OrigFileName);
  End;

  Function TghFileVersionInfo.OrigPath :String;
  Begin
    // Path with original file name

    Result := Str (ghvsOriginalFilename);

    If Result <> '' Then
      Result := ExtractFilePath (Path) + Result
    Else
      Result := Path;
  End;

  Function TghFileVersionInfo.OrigPath (Const Ext :String) :String;
  Begin
    Result := ghExtPath (OrigPath, Ext);
  End;

  Function TghFileVersionInfo.Str (Const TranslationCode :Integer;
    Const Name :String) :PChar;
  Begin
    If Data <> Nil Then
      Result := InternalStr (TranslationCode, Name)
    Else
      Result := Nil;
  End;

  Function TghFileVersionInfo.Str (Const Name :String;
    Const TranslationIndex :Integer = 0) :PChar;
  Begin
    If Data <> Nil Then
      Result := InternalStr (TranslationCodes [TranslationIndex], Name)
    Else
      Result := Nil;
  End;

End.

