{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFObjList.pas - TghObjList class unit.                                 }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFObjList;  { Object List }

Interface

  Uses
    Contnrs, Classes, GHFRTL;

  Type
    { Object List class }
    TghObjList = Class (TObjectList)
      Protected
        { Fields }
        AuxLists :TList;
        ComponentNexus :TComponent;  // TComponentNexus

        { Static methods }
        Function InternalEnum (Const TempList :TghObjList;
          Const Routine :TghEnumProc; Const Params :Variant) :TObject;
        Function InternalAddItems (Const List :TObjectList;
          StartIndex :Integer; Const EndIndex :Integer; Const Cls :TClass)
          :Integer; Overload;
        Function InternalAddItems (Const List :TStrings;
          StartIndex :Integer; Const EndIndex :Integer; Const Cls :TClass)
          :Integer; Overload;

        { Overridden methods }
        Procedure Notify (Ptr :Pointer; Action :TListNotification);
          Override;
      Public
        Destructor Destroy; Override;

        { Static methods }

        Function AddComponents (Const Owner :TComponent;
          Const Cls :TComponentClass = Nil; Const Recurse :Boolean = False)
          :Integer; Overload;
        Function AddComponents (Const Owner :TComponent;
          Const Recurse :Boolean) :Integer; Overload;
        Function AddItems (Const List :TObjectList;
          StartIndex :Integer = 0; Const EndIndex :Integer = MaxInt;
          Const Cls :TClass = Nil) :Integer; Overload;
        Function AddItems (Const List :TObjectList;
          Const StartIndex :Integer; Const Cls :TClass) :Integer; Overload;
        Function AddItems (Const List :TObjectList; Const Cls :TClass)
          :Integer; Overload;
        Function AddItems (Const List :TStrings;
          StartIndex :Integer = 0; Const EndIndex :Integer = MaxInt;
          Const Cls :TClass = Nil) :Integer; Overload;
        Function AddItems (Const List :TStrings;
          Const StartIndex :Integer; Const Cls :TClass) :Integer; Overload;
        Function AddItems (Const List :TStrings; Const Cls :TClass)
          :Integer; Overload;

        // Call methods without Param parameter
        Procedure Call (Const StartIndex, EndIndex :Integer;
          Const Cls :TClass; Const AMethodAddress :Pointer); Overload;
        Procedure Call (Const StartIndex :Integer; Const Cls :TClass;
          Const AMethodAddress :Pointer); Overload;
        Procedure Call (Const StartIndex, EndIndex :Integer;
          Const AMethodAddress :Pointer); Overload;
        Procedure Call (Const Cls :TClass; Const AMethodAddress :Pointer);
          Overload;
        Procedure Call (Const StartIndex :Integer;
          Const AMethodAddress :Pointer); Overload;
        Procedure Call (Const AMethodAddress :Pointer); Overload;

        // Call methods with Param parameter }
        Procedure Call (Const StartIndex, EndIndex :Integer;
          Const Cls :TClass; Const AMethodAddress, Param :Pointer);
          Overload;
        Procedure Call (Const StartIndex :Integer; Const Cls :TClass;
          Const AMethodAddress, Param :Pointer); Overload;
        Procedure Call (Const StartIndex, EndIndex :Integer;
          Const AMethodAddress, Param :Pointer); Overload;
        Procedure Call (Const Cls :TClass;
          Const AMethodAddress, Param :Pointer); Overload;
        Procedure Call (Const StartIndex :Integer;
          Const AMethodAddress, Param :Pointer); Overload;
        Procedure Call (Const AMethodAddress, Param :Pointer); Overload;

        Function Clone (Const StartIndex :Integer; Const Cls :TClass)
          :TghObjList; Overload;
        Function Clone (Const Cls :TClass) :TghObjList; Overload;

        // Enumerate methods with explicit "Params"
        Function Enum (Const StartIndex, EndIndex :Integer;
          Const Cls :TClass; Const Routine :TghEnumProc;
          Const Params :Variant) :TObject; Overload;
        Function Enum (Const StartIndex :Integer; Const Cls :TClass;
          Const Routine :TghEnumProc; Const Params :Variant) :TObject;
          Overload;
        Function Enum (Const StartIndex, EndIndex :Integer;
          Const Routine :TghEnumProc; Const Params :Variant) :TObject;
          Overload;
        Function Enum (Const Cls :TClass; Const Routine :TghEnumProc;
          Const Params :Variant) :TObject; Overload;
        Function Enum (Const StartIndex :Integer;
          Const Routine :TghEnumProc; Const Params :Variant) :TObject;
          Overload;
        Function Enum (Const Routine :TghEnumProc; Const Params :Variant)
          :TObject; Overload;

        // Enumerate methods with implicit (Unassigned) "Params"
        Function Enum (Const StartIndex, EndIndex :Integer;
          Const Cls :TClass; Const Routine :TghEnumProc = Nil) :TObject;
          Overload;
        Function Enum (Const StartIndex :Integer; Const Cls :TClass;
          Const Routine :TghEnumProc = Nil) :TObject; Overload;
        Function Enum (Const StartIndex, EndIndex :Integer;
          Const Routine :TghEnumProc) :TObject; Overload;
        Function Enum (Const Cls :TClass; Const Routine :TghEnumProc = Nil)
          :TObject; Overload;
        Function Enum (Const StartIndex :Integer;
          Const Routine :TghEnumProc) :TObject; Overload;
        Function Enum (Const Routine :TghEnumProc) :TObject; Overload;

        Function Remove (Const Index :Integer) :TObject; Overload;
        
        { Virtual methods }
        Function Clone (StartIndex :Integer = 0;
          EndIndex :Integer = MaxInt; Cls :TClass = Nil) :TghObjList;
          Overload; Virtual;
    End;

    { Object List class type }
    TghObjListClass = Class Of TghObjList;

Implementation

  Uses
    Variants, SysUtils;

  Type
    TComponentNexus = Class (TComponent)
      Protected
        { Fields }
        List :TghObjList;

        { Overridden methods }
        Procedure Notification (AComponent :TComponent;
          Operation :TOperation); Override;
      Public
        Constructor Create (Const AList :TghObjList); Reintroduce;
    End;

  { TComponentNexus }

  Constructor TComponentNexus.Create (Const AList :TghObjList);
  Begin
    Inherited Create (Nil);
    List := AList;
  End;

  { Protected overridden methods }

  Procedure TComponentNexus.Notification (AComponent :TComponent;
    Operation :TOperation);
  Begin
    If Operation = opRemove Then
      List.Extract (AComponent);

    Inherited Notification (AComponent, Operation);
  End;

  { TghObjList }

  Destructor TghObjList.Destroy;
  Begin
    FreeAndNil (AuxLists);
    Inherited Destroy;
    ComponentNexus.Free;
  End;

  { Protected static methods }

  Function TghObjList.InternalEnum (Const TempList :TghObjList;
    Const Routine :TghEnumProc; Const Params :Variant) :TObject;
  Begin
    While TempList.Count > 0 Do
    Begin
      Result := TempList.Remove (0);

      { If the object is still in this (Self) list, we call to Routine.
        NOTE: Routine could destroy any object in the list (see AuxLists
        field). }
      If (IndexOf (Result) > -1) And
      Not ghCall (Routine, Result, Params) Then
        Exit;  // Routine stopped the enumeration
    End;

    Result := Nil;
  End;

  Function TghObjList.InternalAddItems (Const List :TObjectList;
    StartIndex :Integer; Const EndIndex :Integer; Const Cls :TClass)
    :Integer;
  Begin
    Result := 0;

    For StartIndex := StartIndex To EndIndex Do
      If (Cls = Nil) Or (List [StartIndex] Is Cls) Then
      Begin
        Add (List [StartIndex]);
        Inc (Result);
      End;
  End;

  Function TghObjList.InternalAddItems (Const List :TStrings;
    StartIndex :Integer; Const EndIndex :Integer; Const Cls :TClass)
    :Integer;
  Begin
    Result := 0;

    For StartIndex := StartIndex To EndIndex Do
      If (Cls = Nil) Or (List.Objects [StartIndex] Is Cls) Then
      Begin
        Add (List.Objects [StartIndex]);
        Inc (Result);
      End;
  End;

  { Protected overridden methods }

  Procedure TghObjList.Notify (Ptr :Pointer; Action :TListNotification);
  Var
    I :Integer;
  Begin
    If TObject (Ptr) Is TComponent Then
      If Action = lnAdded Then
      Begin
        If ComponentNexus = Nil Then
          ComponentNexus := TComponentNexus.Create (Self);

        TComponent (Ptr).FreeNotification (ComponentNexus);
      End
      Else  // lnExtracted, lnDeleted
        TComponent (Ptr).RemoveFreeNotification (ComponentNexus)
    Else
      If (AuxLists <> Nil) And (Action In [lnExtracted, lnDeleted]) Then
        For I := 0 To AuxLists.Count - 1 Do
          TghObjList (AuxLists [I]).Remove (TObject (Ptr));

    Inherited Notify (Ptr, Action);
  End;

  { Public static methods }

  Function TghObjList.AddComponents (Const Owner :TComponent;
    Const Cls :TComponentClass = Nil; Const Recurse :Boolean = False)
    :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Owner.ComponentCount - 1 Do
    Begin
      If (Cls = Nil) Or (Owner.Components [I] Is Cls) Then
      Begin
        Add (Owner.Components [I]);
        Inc (Result);
      End;

      If Recurse Then
        Inc (Result, AddComponents (Owner.Components [I], Cls, True));
    End;
  End;

  Function TghObjList.AddComponents (Const Owner :TComponent;
    Const Recurse :Boolean) :Integer;
  Begin
    Result := AddComponents (Owner, TComponentClass (Nil), Recurse);
  End;

  Function TghObjList.AddItems (Const List :TObjectList;
    StartIndex :Integer = 0; Const EndIndex :Integer = MaxInt;
    Const Cls :TClass = Nil) :Integer;
  Begin
    Result := InternalAddItems (List, ghIndexMin (StartIndex),
      ghIndexMax (EndIndex, List.Count), Cls);
  End;

  Function TghObjList.AddItems (Const List :TObjectList;
    Const StartIndex :Integer; Const Cls :TClass) :Integer;
  Begin
    Result := InternalAddItems (List, ghIndexMin (StartIndex),
      List.Count - 1, Cls);
  End;

  Function TghObjList.AddItems (Const List :TObjectList; Const Cls :TClass)
    :Integer;
  Begin
    Result := InternalAddItems (List, 0, List.Count - 1, Cls);
  End;

  Function TghObjList.AddItems (Const List :TStrings;
    StartIndex :Integer = 0; Const EndIndex :Integer = MaxInt;
    Const Cls :TClass = Nil) :Integer;
  Begin
    Result := InternalAddItems (List, ghIndexMin (StartIndex),
      ghIndexMax (EndIndex, List.Count), Cls);
  End;

  Function TghObjList.AddItems (Const List :TStrings;
    Const StartIndex :Integer; Const Cls :TClass) :Integer;
  Begin
    Result := InternalAddItems (List, ghIndexMin (StartIndex),
      List.Count - 1, Cls);
  End;

  Function TghObjList.AddItems (Const List :TStrings; Const Cls :TClass)
    :Integer;
  Begin
    Result := InternalAddItems (List, 0, List.Count - 1, Cls);
  End;

  Procedure TghObjList.Call (Const StartIndex, EndIndex :Integer;
    Const Cls :TClass; Const AMethodAddress :Pointer);

    // Call back routine
    Procedure CallMethod (Const Obj :TObject;
      Const AMethodAddress :Variant; Var AContinue :Boolean);
    Begin
      ghCall (Obj, ghPtr (AMethodAddress));
    End;
  Begin
    Enum (StartIndex, EndIndex, Cls, @CallMethod, ghVar (AMethodAddress));
  End;

  Procedure TghObjList.Call (Const StartIndex :Integer; Const Cls :TClass;
    Const AMethodAddress :Pointer);
  Begin
    Call (StartIndex, Count - 1, Cls, AMethodAddress);
  End;

  Procedure TghObjList.Call (Const StartIndex, EndIndex :Integer;
    Const AMethodAddress :Pointer);
  Begin
    Call (StartIndex, EndIndex, TClass (Nil), AMethodAddress);
  End;

  Procedure TghObjList.Call (Const Cls :TClass;
    Const AMethodAddress :Pointer);
  Begin
    Call (0, Count - 1, Cls, AMethodAddress);
  End;

  Procedure TghObjList.Call (Const StartIndex :Integer;
    Const AMethodAddress :Pointer);
  Begin
    Call (StartIndex, Count - 1, TClass (Nil), AMethodAddress);
  End;

  Procedure TghObjList.Call (Const AMethodAddress :Pointer);
  Begin
    Call (0, Count - 1, TClass (Nil), AMethodAddress);
  End;

  Procedure TghObjList.Call (Const StartIndex, EndIndex :Integer;
    Const Cls :TClass; Const AMethodAddress, Param :Pointer);

    // Call back routine
    Procedure CallMethod (Const Obj :TObject; Const Params :Variant;
      Var Continue :Boolean);
    Begin
      ghCall (Obj, ghPtr (Params [0]), ghPtr (Params [1]));
    End;
  Begin
    Enum (StartIndex, EndIndex, Cls, @CallMethod,
      ghVar ([AMethodAddress, Param]));
  End;

  Procedure TghObjList.Call (Const StartIndex :Integer; Const Cls :TClass;
    Const AMethodAddress, Param :Pointer);
  Begin
    Call (StartIndex, Count - 1, Cls, AMethodAddress, Param);
  End;

  Procedure TghObjList.Call (Const StartIndex, EndIndex :Integer;
    Const AMethodAddress, Param :Pointer);
  Begin
    Call (StartIndex, EndIndex, TClass (Nil), AMethodAddress, Param);
  End;

  Procedure TghObjList.Call (Const Cls :TClass;
    Const AMethodAddress, Param :Pointer);
  Begin
    Call (0, Count - 1, Cls, AMethodAddress, Param);
  End;

  Procedure TghObjList.Call (Const StartIndex :Integer;
    Const AMethodAddress, Param :Pointer);
  Begin
    Call (StartIndex, Count - 1, TClass (Nil), AMethodAddress, Param);
  End;

  Procedure TghObjList.Call (Const AMethodAddress, Param :Pointer);
  Begin
    Call (0, Count - 1, TClass (Nil), AMethodAddress, Param);
  End;

  Function TghObjList.Clone (Const StartIndex :Integer; Const Cls :TClass)
    :TghObjList;
  Begin
    Result := Clone (StartIndex, Count - 1, Cls);
  End;

  Function TghObjList.Clone (Const Cls :TClass) :TghObjList;
  Begin
    Result := Clone (0, Count - 1, Cls);
  End;

  Function TghObjList.Enum (Const StartIndex, EndIndex :Integer;
    Const Cls :TClass; Const Routine :TghEnumProc; Const Params :Variant)
    :TObject;
  Var
    TempList :TghObjList;
  Begin
    { NOTE: We use a temporary list (snapshot of specific sublist) because
      Routine could alter the content of this list. }
    TempList := TghObjList.Create (False);

    Try
      If TempList.AddItems (Self, StartIndex, EndIndex, Cls) > 0 Then
      Begin
        If AuxLists = Nil Then
          AuxLists := TList.Create;

        AuxLists.Add (TempList);

        Try
          Result := InternalEnum (TempList, Routine, Params);
        Finally
          AuxLists.Remove (TempList);
        End;
      End
      Else
        Result := Nil;
    Finally
      TempList.Free;
    End;
  End;

  Function TghObjList.Enum (Const StartIndex :Integer; Const Cls :TClass;
    Const Routine :TghEnumProc; Const Params :Variant) :TObject;
  Begin
    Result := Enum (StartIndex, Count - 1, Cls, Routine, Params);
  End;

  Function TghObjList.Enum (Const StartIndex, EndIndex :Integer;
    Const Routine :TghEnumProc; Const Params :Variant) :TObject;
  Begin
    Result := Enum (StartIndex, EndIndex, TClass (Nil), Routine, Params);
  End;

  Function TghObjList.Enum (Const Cls :TClass; Const Routine :TghEnumProc;
    Const Params :Variant) :TObject;
  Begin
    Result := Enum (0, Count - 1, Cls, Routine, Params);
  End;

  Function TghObjList.Enum (Const StartIndex :Integer;
    Const Routine :TghEnumProc; Const Params :Variant) :TObject;
  Begin
    Result := Enum (StartIndex, Count - 1, TClass (Nil), Routine, Params);
  End;

  Function TghObjList.Enum (Const Routine :TghEnumProc;
    Const Params :Variant) :TObject;
  Begin
    Result := Enum (0, Count - 1, TClass (Nil), Routine, Params);
  End;

  Function TghObjList.Enum (Const StartIndex, EndIndex :Integer;
    Const Cls :TClass; Const Routine :TghEnumProc = Nil) :TObject;
  Begin
    Result := Enum (StartIndex, EndIndex, Cls, Routine, Unassigned);
  End;

  Function TghObjList.Enum (Const StartIndex :Integer; Const Cls :TClass;
    Const Routine :TghEnumProc = Nil) :TObject;
  Begin
    Result := Enum (StartIndex, Count - 1, Cls, Routine, Unassigned);
  End;

  Function TghObjList.Enum (Const StartIndex, EndIndex :Integer;
    Const Routine :TghEnumProc) :TObject;
  Begin
    Result := Enum (StartIndex, EndIndex, TClass (Nil), Routine,
      Unassigned);
  End;

  Function TghObjList.Enum (Const Cls :TClass;
    Const Routine :TghEnumProc = Nil) :TObject;
  Begin
    Result := Enum (0, Count - 1, Cls, Routine, Unassigned);
  End;

  Function TghObjList.Enum (Const StartIndex :Integer;
    Const Routine :TghEnumProc) :TObject;
  Begin
    Result := Enum (StartIndex, Count - 1, TClass (Nil), Routine,
      Unassigned);
  End;

  Function TghObjList.Enum (Const Routine :TghEnumProc) :TObject;
  Begin
    Result := Enum (0, Count - 1, TClass (Nil), Routine, Unassigned);
  End;

  Function TghObjList.Remove (Const Index :Integer) :TObject;
  Begin
    Result := Items [Index];
    Delete (Index);
  End;

  { Public virtual methods }

  Function TghObjList.Clone (StartIndex :Integer = 0;
    EndIndex :Integer = MaxInt; Cls :TClass = Nil) :TghObjList;
  Begin
    Result := TghObjListClass (ClassType).Create (False);

    Try
      Result.AddItems (Self, StartIndex, EndIndex, Cls);
    Except
      Result.Free;
      Raise;
    End;
  End;

End.

