{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFSQLConnection.pas - TghSQLConnection class unit.                     }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFSQLConnection;  { SQL Connection }

Interface

  Uses
    GHFCustomSavePoint, GHFCustomPurposeSavePoints, SQLExpr, DBXpress,
    Classes, GHFVCL, GHFRTL;

  Type
    { SQL Connection Save Point class }
    TghSQLConnectionSavePoint = Class (TghCustomSavePoint)
      Protected
        { Fields }
        Name :String;

        { Overridden methods }
        Procedure InternalConfirm; Override;
        Procedure InternalRelease; Override;
        Procedure InternalRestore; Override;
        Procedure InternalSave; Override;
      Public
        Constructor Create (AOwner :TghCustomSavePoints); Override;
    End;

    { SQL Connection Transaction Save Point class }
    TghSQLConnectionTransSavePoint = Class (TghCustomSavePoint)
      Protected
        { Overridden methods }
        Procedure InternalConfirm; Override;
        Procedure InternalRestore; Override;
        Procedure InternalSave; Override;
    End;

    { SQL Connection Save Points class }
    TghSQLConnectionSavePoints = Class (TghCustomPurposeSavePoints)
      Protected
        { Overridden methods }
        Function NewPointClass :TghCustomSavePointClass; Override;
    End;

    { TghSQLConnectionSavePoints class type }
    TghSQLConnectionSavePointsClass = Class Of
      TghSQLConnectionSavePoints;

    { SQL Connection class }
    TghSQLConnection = Class (TSQLConnection)
      Protected
        { Fields }
        FNameCase :TghChrCase;
        FSavePoints :TghSQLConnectionSavePoints;
        FStoreConnected :Boolean;

        { Virtual methods }

        { Current Date Expression }
        Function CurrentDateExpr :String; Virtual;

        { Current Time Expression }
        Function CurrentTimeExpr :String; Virtual;

        { Current Time Stamp Expression }
        Function CurrentTimeStampExpr :String; Virtual;

        { Next Value Expression }
        Function NextValueExpr :String; Virtual;

        Function ReleaseSavePointSyntax :String; Virtual;
        Function RollbackSavePointSyntax :String; Virtual;
        Function SavePointsClass :TghSQLConnectionSavePointsClass;
          Virtual;
        Function SavePointSyntax :String; Virtual;

        { Select Auxiliary Table }
        Function SelectAuxTable :String; Virtual;

        { Overridden methods }
        Procedure DoDisconnect; Override;
      Public
        { Fields }
        Transaction :PTTransactionDesc;

        Constructor Create (AOwner :TComponent); Override;
        Destructor Destroy; Override;

        { Static methods }
        Function CasedName (Const AName :String) :String;
        Procedure Commit (Const RestartTransaction :Boolean = False);
          Overload;
        Function CurrentDate :TDateTime;
        Function CurrentTime :TDateTime;
        Function CurrentTimeStamp :TDateTime;
        Procedure Delete (Const Table :String; Const ID :Integer);
        Function ExecProc (Const Proc :String; Const AParams :Variant)
          :Variant;
        Function ExecProcStr (Const Proc :String; Const AParams :Variant)
          :String;
        Procedure Insert (Const Table, Fields :String;
          Const Values :Variant); Overload;
        Procedure Insert (Const Table, Fields :String;
          Const Value1, Value2 :Integer); Overload;
        Function Max (Const Table, Field :String;
          Const Conditions :String = '') :Variant;
        Function MaxInc (Const Table, Field :String;
          Const Conditions :String = ''; Const Inc :Integer = 1) :Variant;
          Overload;
        Function MaxInc (Const Table, Field :String; Const Inc :Integer)
          :Variant; Overload;
        Function Min (Const Table, Field :String;
          Const Conditions :String = '') :Variant;
        Function NextValue (Const Sequence :String) :Integer;
        Function QuotedCasedName (Const AName :String) :String;
        Procedure Rollback (Const RestartTransaction :Boolean = False);
          Overload;
        Function Select (Const Statement :String) :Variant;
        Function SelectConfig (Const Fields :String) :Variant;
        Function SelectExpr (Const Expr :String) :Variant;
        Function SelectFloat (Const Statement :String) :Extended;
        Function SelectID (Const SubStatement :String) :Integer;
        Function SelectInt (Const Statement :String) :Integer;
        Function SelectStr (Const Statement :String) :String;
        Function SelectTrim (Const Statement :String) :String;
        Procedure StartTransaction; Overload;

        { Virtual methods }
        Function ConstraintField (Const Constraint :String) :String;
          Virtual;
        Function ConstraintRefTable (Const Constraint :String) :String;
          Virtual;
        Function ConstraintTable (Const Constraint :String) :String;
          Virtual;
        Function FieldConstraint (Const Table, Field :String) :String;
          Virtual;
        Function FieldOfIndex (Const Index :String) :String; Virtual;
        Function FieldRefTable (Const Table, Field :String) :String;
          Virtual;
        Function PrimaryKeyIndex (Const Table :String) :String; Virtual;
        Function TableError (Const Msg :String) :TghTableError;
          Virtual;
        Function TableOfIndex (Const Index :String) :String; Virtual;
      Published
        { Properties }
        Property Connected Stored FStoreConnected;
        Property NameCase :TghChrCase Read FNameCase Write FNameCase
          Default ghcaDefault;
        Property SavePoints :TghSQLConnectionSavePoints
          Read FSavePoints;
        Property StoreConnected :Boolean Read FStoreConnected
          Write FStoreConnected Default False;
    End;

Implementation

  Uses
    SysUtils, Variants, StrUtils;

  { TghSQLConnectionSavePoint }

  Constructor TghSQLConnectionSavePoint.Create (
    AOwner :TghCustomSavePoints);
  Begin
    If AOwner.Last Is TghSQLConnectionSavePoint Then
      Name := ghInc (TghSQLConnectionSavePoint (AOwner.Last).Name)
    Else
      Name := 'sp1';

    Inherited Create (AOwner);
  End;

  { Protected overridden methods }

  Procedure TghSQLConnectionSavePoint.InternalConfirm;
  Begin
    InternalRelease;  { Delete this and following save points }
  End;

  Procedure TghSQLConnectionSavePoint.InternalRelease;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      ExecuteDirect (Format (ReleaseSavePointSyntax (), [Self.Name]));
  End;

  Procedure TghSQLConnectionSavePoint.InternalRestore;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      ExecuteDirect (Format (RollbackSavePointSyntax (), [Self.Name]));
  End;

  Procedure TghSQLConnectionSavePoint.InternalSave;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      ExecuteDirect (Format (SavePointSyntax (), [Self.Name]));
  End;

  { TghSQLConnectionTransSavePoint }

  { Protected overridden methods }

  Procedure TghSQLConnectionTransSavePoint.InternalConfirm;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      Commit (Transaction^);
  End;

  Procedure TghSQLConnectionTransSavePoint.InternalRestore;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      Rollback (Transaction^);

    Invalidate;
  End;

  Procedure TghSQLConnectionTransSavePoint.InternalSave;
  Begin
    With TghSQLConnection (Owner.Owner) Do
      StartTransaction (Transaction^);
  End;

  { TghSQLConnectionSavePoints }

  { Protected overridden methods }

  Function TghSQLConnectionSavePoints.NewPointClass
    :TghCustomSavePointClass;
  Begin
    If Count = 0 Then
      { First point is the start of transaction }
      Result := TghSQLConnectionTransSavePoint
    Else
      { Save points inside of transaction }
      Result := TghSQLConnectionSavePoint;
  End;

  { TghSQLConnection }

  Constructor TghSQLConnection.Create (AOwner :TComponent);
  Begin
    Inherited Create (AOwner);
    FSavePoints := SavePointsClass ().Create (Self);
    ghSetSubcomponent (SavePoints, 'SavePoints');

    If Not (csDesigning In ComponentState) Then
    Begin
      New (Transaction);
      Transaction.TransactionID := 1;
      Transaction.IsolationLevel := xilReadCommitted;
    End;
  End;

  Destructor TghSQLConnection.Destroy;
  Begin
    FreeAndNil (FSavePoints);
    
    If Not (csDesigning In ComponentState) Then
    Begin
      Dispose (Transaction);
      Transaction := Nil;
    End;

    Inherited Destroy;
  End;

  { Protected static methods }

  { Protected virtual methods }

  Function TghSQLConnection.CurrentDateExpr :String;
  Begin
    Result := Params.Values ['CurrentDateExpr'];

    If Result = '' Then
      Result := 'Current_Date';
  End;

  Function TghSQLConnection.CurrentTimeExpr :String;
  Begin
    Result := Params.Values ['CurrentTimeExpr'];

    If Result = '' Then
      Result := 'Current_Time';
  End;

  Function TghSQLConnection.CurrentTimeStampExpr :String;
  Begin
    Result := Params.Values ['CurrentTimeStampExpr'];

    If Result = '' Then
      Result := 'Current_TimeStamp';
  End;

  Function TghSQLConnection.NextValueExpr :String;
  Begin
    Result := Params.Values ['NextValueExpr'];

    If Result = '' Then
      Result := 'Next Value For %s';
  End;

  Function TghSQLConnection.ReleaseSavePointSyntax :String;
  Begin
    Result := Params.Values ['ReleaseSavePointSyntax'];

    If Result = '' Then
      Result := 'Release SavePoint %s';
  End;

  Function TghSQLConnection.RollbackSavePointSyntax :String;
  Begin
    Result := Params.Values ['RollbackSavePointSyntax'];

    If Result = '' Then
      Result := 'Rollback To SavePoint %s';
  End;

  Function TghSQLConnection.SavePointsClass
    :TghSQLConnectionSavePointsClass;
  Begin
    Result := TghSQLConnectionSavePoints;
  End;

  Function TghSQLConnection.SavePointSyntax :String;
  Begin
    Result := Params.Values ['SavePointSyntax'];

    If Result = '' Then
      Result := 'SavePoint %s';
  End;

  Function TghSQLConnection.SelectAuxTable :String;
  Begin
    Result := CasedName (Params.Values ['SelectAuxTable']);
  End;

  Procedure TghSQLConnection.DoDisconnect;
  Begin
    Inherited DoDisconnect;

    // Save points are freed in the destructor
    If Not (csDestroying In ComponentState) Then
      SavePoints.Invalidate;
  End;

  { Public static methods }

  Function TghSQLConnection.CasedName (Const AName :String) :String;
  Begin
    Result := ghCased (AName, NameCase);
  End;

  Procedure TghSQLConnection.Commit (
    Const RestartTransaction :Boolean = False);
  Begin
    SavePoints.First.Confirm (RestartTransaction);
  End;

  Function TghSQLConnection.CurrentDate :TDateTime;
  Begin
    Result := SelectExpr (CurrentDateExpr ());
  End;

  Function TghSQLConnection.CurrentTime :TDateTime;
  Begin
    Result := SelectExpr (CurrentTimeExpr ());
  End;

  Function TghSQLConnection.CurrentTimeStamp :TDateTime;
  Begin
    Result := SelectExpr (CurrentTimeStampExpr ());
  End;

  Procedure TghSQLConnection.Delete (Const Table :String;
    Const ID :Integer);
  Begin
    ExecuteDirect ('Delete From ' + CasedName (Table) + ' Where ID = ' +
      IntToStr (ID));
  End;

  Function TghSQLConnection.ExecProc (Const Proc :String;
    Const AParams :Variant) :Variant;
  Var
    I :Integer;
  Begin
    With TSQLStoredProc.Create (Self) Do
      Try
        SQLConnection := Self;
        StoredProcName := CasedName (Proc);

        { Assign input parameters }
        If VarIsArray (AParams) Then
          For I := 0 To ghHigh (AParams) Do
            ghInputParam (Params, I).Value := AParams [I]
        Else
          ghInputParam (Params).Value := AParams;

        ExecProc;
        I := ghOutputParamCount (Params);

        { Return output parameter(s) or Null }
        Case I Of
          0 : Result := Null;
          1 : Result := ghOutputParam (Params).Value;
          Else
          Begin
            Result := ghVarArr (I);

            For I := 0 To I - 1 Do
              Result [I] := ghOutputParam (Params, I).Value;
          End
        End;
      Finally
        Free;
      End;
  End;

  Function TghSQLConnection.ExecProcStr (Const Proc :String;
    Const AParams :Variant) :String;
  Begin
    Result := VarToStr (ExecProc (Proc, AParams));
  End;

  Procedure TghSQLConnection.Insert (Const Table, Fields :String;
    Const Values :Variant);
  Begin
    ExecuteDirect ('Insert Into ' + CasedName (Table) + ' (' +
      ANSIReplaceStr (Fields, ';', ', ') + ') Values (' +
      ghCommaSQLValues (Values) + ')');
  End;

  Procedure TghSQLConnection.Insert (Const Table, Fields :String;
    Const Value1, Value2 :Integer);
  Begin
    Insert (Table, Fields, VarArrayOf ([Value1, Value2]));
  End;

  Function TghSQLConnection.Max (Const Table, Field :String;
    Const Conditions :String = '') :Variant;
  Begin
    Result := Select ('Max (' + CasedName (Field) + ') From ' +
      CasedName (Table) + ghPrefixed (Conditions, ' Where '));

    If VarIsNull (Result) Then
      Result := 0;
  End;

  Function TghSQLConnection.MaxInc (Const Table, Field :String;
    Const Conditions :String = ''; Const Inc :Integer = 1) :Variant;
  Begin
    Result := ghInc (Max (Table, Field, Conditions), Inc);
  End;

  Function TghSQLConnection.MaxInc (Const Table, Field :String;
    Const Inc :Integer) :Variant;
  Begin
    Result := MaxInc (Table, Field, '', Inc);
  End;

  Function TghSQLConnection.Min (Const Table, Field :String;
    Const Conditions :String = '') :Variant;
  Begin
    Result := Select ('Min (' + CasedName (Field) + ') From ' +
      CasedName (Table) + ghPrefixed (Conditions, ' Where '));

    If VarIsNull (Result) Then
      Result := 0;
  End;

  Function TghSQLConnection.NextValue (Const Sequence :String) :Integer;
  Begin
    Result := SelectExpr (Format (NextValueExpr (), [Sequence]));
  End;

  Function TghSQLConnection.QuotedCasedName (Const AName :String)
    :String;
  Begin
    Result := QuotedStr (CasedName (AName));
  End;

  Procedure TghSQLConnection.Rollback (
    Const RestartTransaction :Boolean = False);
  Begin
    SavePoints.First.Restore (RestartTransaction);
  End;

  Function TghSQLConnection.Select (Const Statement :String) :Variant;
  Var
    ResultSet :TCustomSQLDataSet;
    I :Integer;
  Begin
    Execute (ghPrefixed (Statement, 'Select '), Nil, @ResultSet);

    With ResultSet Do
      Try
        If FieldCount = 1 Then
          Result := Fields [0].Value
        Else
        Begin
          Result := VarArrayCreate([0, FieldCount - 1], varVariant);

          For I := 0 To FieldCount - 1 Do
            Result [I] := Fields [I].Value;
        End;
      Finally
        Free;
      End;
  End;

  Function TghSQLConnection.SelectConfig (Const Fields :String)
    :Variant;
  Begin
    Result := Select (CasedName (Fields) + ' From ' +
      CasedName (Params.Values ['ConfigTable']));
  End;

  Function TghSQLConnection.SelectExpr (Const Expr :String)
    :Variant;
  Begin
    Result := Select (Expr + ghPrefixed (SelectAuxTable, ' From '));
  End;

  Function TghSQLConnection.SelectFloat (Const Statement :String)
    :Extended;
  Begin
    Result := ghFloat (Select (Statement));
  End;

  Function TghSQLConnection.SelectID (Const SubStatement :String) :Integer;
  Begin
    Result := SelectInt ('ID ' + ghPrefixed (SubStatement, 'From '));
  End;

  Function TghSQLConnection.SelectInt (Const Statement :String)
    :Integer;
  Begin
    Result := ghInt (Select (Statement));
  End;

  Function TghSQLConnection.SelectStr (Const Statement :String) :String;
  Begin
    Result := VarToStr (Select (Statement));
  End;

  Function TghSQLConnection.SelectTrim (Const Statement :String) :String;
  Begin
    Result := Trim (SelectStr (Statement));
  End;

  Procedure TghSQLConnection.StartTransaction;
  Begin
    If SavePoints.Count = 0 Then
      SavePoints.Save
    Else
      SavePoints.First.Save;
  End;

  { Public virtual methods }

  Function TghSQLConnection.ConstraintField (Const Constraint :String)
    :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.ConstraintRefTable (Const Constraint :String)
    :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.ConstraintTable (Const Constraint :String)
    :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.FieldConstraint (Const Table, Field :String)
    :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.FieldOfIndex (Const Index :String) :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.FieldRefTable (Const Table, Field :String)
    :String;
  Begin
    Result := ConstraintRefTable (FieldConstraint (Table, Field));
  End;

  Function TghSQLConnection.PrimaryKeyIndex (Const Table :String) :String;
  Begin
    Result := '';
  End;

  Function TghSQLConnection.TableError (Const Msg :String) :TghTableError;
  Begin
    Result.Code := 0;
    Result.Table := '';
    Result.Field := '';
  End;

  Function TghSQLConnection.TableOfIndex (Const Index :String) :String;
  Begin
    Result := '';
  End;

End.

