{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFUtils.pas - Utility element unit.                                    }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFUtils;  { Utilities }

Interface

  Uses
    GHFEN, GHFRTL, Classes, GHFCustomSavePoint, GHFStrList,
    GHFFileVersionInfo, IniFiles, SysUtils, Controls, DB;

  Const
    { Pascal Characters (of identifiers) }
    ghpcAlphanum = ghltENAll + ghdgDecimal;  // Alphanumeric
    ghpcInitial = ghltENAll + ['_'];
    ghpcAll = ghpcAlphanum + ['_'];

  Var
    GHTransSavePoints :TghCustomSavePoints;

  { Functions and procedures }

  { Confirm Save Point }
  Procedure ghConfirmSavePoint (Var SavePoint);

  { Enable }
  Procedure ghEnable (Const Controls :Array Of TControl;
    Const Enable :Boolean = True; Const Recurse :Boolean = False);

  { Enumerate Fields }
  Function ghEnumFields (Const DataSet :TDataSet;
    Const Cls :TFieldClass; Const Routine :TghEnumProc;
    Const Params :Variant) :TField; Overload;

  { Enumerate Fields }
  Function ghEnumFields (Const DataSet :TDataSet;
    Const Routine :TghEnumProc; Const Params :Variant) :TField; Overload;

  { Executable Version Information }
  Function ghExeVersionInfo :TghFileVersionInfo;

  { Fix Up Providers }
  Procedure ghFixUpProviders (
    Const DataSetOwner, ProviderOwner :TComponent);

  { Ini Exists? }
  Function ghIniExists :Boolean;

  { Ini File }
  Function ghIniFile :TIniFile;

  { Ini Path }
  Function ghIniPath :String;

  { Log }
  Procedure ghLog (Str :String); Overload;

  { Log }
  Procedure ghLog (Const E :Exception); Overload;

  { Log Path }
  Function ghLogPath :String;

  { Parameter }
  Function ghParam (Const Str, Name :String; Const Separator :String = ';')
    :String;

  { Restart Application }
  Procedure ghRestartApp;

  { Restore Save Point }
  Procedure ghRestoreSavePoint (Var SavePoint);

  { Split List }
  Function ghSplitList (Const Value :String; Const Separator :String = ' ';
    Const IncludeEmpties :Boolean = False) :TghStrList; Overload;

  { Split List }
  Function ghSplitList (Const Value :String; Const IncludeEmpties :Boolean)
    :TghStrList; Overload;

  { String List }
  Function ghStrList (Const Strs :Array Of String) :TghStrList;

Implementation

  Uses
    GHFClientDataSet, Provider, GHFObjList, Windows, Forms, GHFObjListEx;

  Type
    TExeVersionInfo = Class (TghFileVersionInfo)
      Public
        Destructor Destroy; Override;
    End;

    TIniFile = Class (IniFiles.TIniFile)
      Public
        Destructor Destroy; Override;
    End;

  Var
    ExeVersionInfo :TExeVersionInfo;
    IniFile :TIniFile;
    RestartApp :Boolean;

  { TExeVersionInfo }

  Destructor TExeVersionInfo.Destroy;
  Begin
    Inherited Destroy;

    If Self = ExeVersionInfo Then
      ExeVersionInfo := Nil;
  End;

  { TIniFile }

  Destructor TIniFile.Destroy;
  Begin
    Inherited Destroy;

    If Self = IniFile Then
      IniFile := Nil;
  End;

  { Functions and procedures }

  Procedure ghConfirmSavePoint (Var SavePoint);
  Begin
    TghCustomSavePoint (SavePoint).Confirm;
    Pointer (SavePoint) := Nil;
  End;

  Procedure ghEnable (Const Controls :Array Of TControl;
    Const Enable :Boolean = True; Const Recurse :Boolean = False);
  Var
    I :Integer;
    List :TghObjListEx;

    // Call back routine
    Procedure SetEnabled (Const Control :TControl; Const Enable :Variant;
      Var Continue :Boolean);
    Begin
      Control.Enabled := Enable;
    End;
  Begin
    List := TghObjListEx.Create (False);

    Try
      For I := 0 To High (Controls) Do
      Begin
        List.Add (Controls [I]);

        If Recurse And (Controls [I] Is TWinControl) Then
          List.AddControls (TWinControl (Controls [I]), True);
      End;

      List.Enum (@SetEnabled, Enable);
    Finally
      List.Free;
    End;
  End;

  Function ghEnumFields (Const DataSet :TDataSet;
    Const Cls :TFieldClass; Const Routine :TghEnumProc;
    Const Params :Variant) :TField; Overload;
  Begin
    With TghObjList.Create (False) Do
      Try
        AddItems (DataSet.FieldList, Cls);
        Result := Enum (Routine, Params) As TField;
      Finally
        Free;
      End;
  End;

  Function ghEnumFields (Const DataSet :TDataSet;
    Const Routine :TghEnumProc; Const Params :Variant) :TField; Overload;
  Begin
    Result := ghEnumFields (DataSet, TField, Routine, Params);
  End;

  Function ghExeVersionInfo :TghFileVersionInfo;
  Begin
    If ExeVersionInfo = Nil Then
      ExeVersionInfo := TExeVersionInfo.Create (MainInstance);

    Result := ExeVersionInfo
  End;

  {}//probar
  { NOTE: This function helps to solve a design time bug of Delphi 7
    ("external reference in inherited component"), when using the Provider
    property of TghClientDataSet objects placed in inherited containers. }
  Procedure ghFixUpProviders (
    Const DataSetOwner, ProviderOwner :TComponent);

    // Call back routine
    Procedure FixUp (Const DataSet :TghClientDataSet;
      Const ProviderOwner :Variant; Var Continue :Boolean);
    Begin
      If (DataSet.ProviderName <> '') And (DataSet.Provider = Nil) Then
        DataSet.Provider := TComponent (
          ghPtr (ProviderOwner)).FindComponent (DataSet.ProviderName) As
          TCustomProvider;
    End;
  Begin
    With TghObjList.Create (False) Do
      Try
        AddComponents (DataSetOwner, TghClientDataSet);
        Enum (@FixUp, ghVar (ProviderOwner));
      Finally
        Free;
      End;
  End;

  Function ghIniExists :Boolean;
  Begin
    Result := FileExists (ghIniPath);
  End;
  
  Function ghIniFile :IniFiles.TIniFile;
  Begin
    If IniFile = Nil Then
      IniFile := TIniFile.Create (ghIniPath);

    Result := IniFile;
  End;

  Function ghIniPath :String;
  Begin
    Result := ghExeVersionInfo.OrigPath ('ini');
  End;

  Procedure ghLog (Str :String); Overload;
  Var
    OpenMode :Word;
  Begin
    If FileExists (ghLogPath) Then
      OpenMode := fmOpenReadWrite Or fmShareDenyNone
    Else
      OpenMode := fmCreate;

    With TFileStream.Create (ghLogPath, OpenMode) Do
      Try
        Seek (0, soFromEnd);
        Str := FormatDateTime ('yyyy-mm-dd hh:nn:ss', Now ()) + ghccCRLF +
          Str + ghccTwoCRLF;
        Write (Str [1], Length (Str));
      Finally
        Free;
      End;
  End;

  Procedure ghLog (Const E :Exception); Overload;
  Begin
    ghLog (E.ClassName () + ': ' + E.Message);
  End;

  Function ghLogPath :String;
  Begin
    Result := ghExeVersionInfo.OrigPath ('log');
  End;

  {}//probar
  Function ghParam (Const Str, Name :String; Const Separator :String = ';')
    :String;
  Begin
    If Str <> '' Then
      With ghSplitList (Str, Separator) Do
        Try
          Result := Values [Name];
        Finally
          Free;
        End
    Else
      Result := '';
  End;

  Procedure ghRestartApp;
  Begin
    Application.Terminate;
    RestartApp := True;
  End;

  Procedure ghRestoreSavePoint (Var SavePoint);
  Begin
    TghCustomSavePoint (SavePoint).Restore;
    Pointer (SavePoint) := Nil;
  End;

  Function ghSplitList (Const Value :String; Const Separator :String = ' ';
    Const IncludeEmpties :Boolean = False) :TghStrList; Overload;
  Begin
    Result := ghStrList (ghSplit (Value, Separator, IncludeEmpties));
  End;

  Function ghSplitList (Const Value :String; Const IncludeEmpties :Boolean)
    :TghStrList; Overload;
  Begin
    Result := ghSplitList (Value, ' ', IncludeEmpties);
  End;

  Function ghStrList (Const Strs :Array Of String) :TghStrList;
  Var
    I :Integer;
  Begin
    Result := TghStrList.Create;

    With Result Do
      Try
        Capacity := Length (Strs);

        For I := 0 To High (Strs) Do
          Add (Strs [I]);
      Except
        Free;
        Raise;
      End;
  End;

Initialization

Finalization
  ExeVersionInfo.Free;
  IniFile.Free;

  If RestartApp Then
    ghExecute (ghExe, ghParams);

End.

