{}// Beta version
{*************************************************************************}
{ GH Freebrary                                                            }
{                                                                         }
{ GHFVCL.pas - Unit of VCL related elements.                              }
{                                                                         }
{ This software is distributed under the open source BSD license.         }
{                                                                         }
{ Copyright (c) 2002-2013, Al Gonzalez (rescatandoadelphi.blogspot.com)   }
{ All rights reserved.                                                    }
{                                                                         }
{ Redistribution and use in source and binary forms, with or without      }
{ modification, are permitted provided that the following conditions are  }
{ met:                                                                    }
{                                                                         }
{   - Redistributions of source code must retain the above copyright      }
{     notice, this list of conditions and the following disclaimer.       }
{   - Redistributions in binary form must reproduce the above copyright   }
{     notice, this list of conditions and the following disclaimer in the }
{     documentation and/or other materials provided with the              }
{     distribution.                                                       }
{   - Neither the name of Al Gonzalez nor the names of his contributors   }
{     may be used to endorse or promote products derived from this        }
{     software without specific prior written permission.                 }
{                                                                         }
{ THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS }
{ IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED   }
{ TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A         }
{ PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT      }
{ HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,  }
{ SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT        }
{ LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,   }
{ DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY   }
{ THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT     }
{ (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE   }
{ OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.    }
{*************************************************************************}

{ Please contribute to improve this library.  I am not a good programmer,
  but good programmers know what Delphi is worth.

  A.G. }

Unit GHFVCL;  { VCL }

Interface

  Uses
    DB, Classes, Controls, DSIntf, GHFRTL, StdCtrls, Windows;

  Const
    { DataSnap }

    { Error codes.  This constants are based on native error codes used
      internally by DataSnap. }
    ghdaErrorBaseIntegrity = $2600;
    ghdaErrorCodeKeyViol = 1;
    ghdaErrorKeyViol = ghdaErrorBaseIntegrity + ghdaErrorCodeKeyViol;
    ghdaErrorBaseInvalidReq = $2700;
    ghdaErrorCodeOutOfRange = 1;
    ghdaErrorOutOfRange = ghdaErrorBaseInvalidReq +
      ghdaErrorCodeOutOfRange;

    // Record Attributes
    // Changes
    ghdaRecAttrsChanges = dsRecDeleted Or dsRecNew Or dsRecModified;

    { TDataSetState sets }
    ghdsBrowsing = [dsBrowse, dsBlockRead];
    ghdsCalc = [dsCalcFields, dsInternalCalc];  { Calculation }
    ghdsEditing = [dsEdit, dsInsert];
    ghdsBasics  = [dsInactive, dsBrowse] + ghdsEditing;
    ghdsBrowsingEditing  = ghdsBrowsing + ghdsEditing;
    ghdsBrowsingEditingCalc  = ghdsBrowsingEditing + ghdsCalc;
    ghdsEditCalc = [dsEdit] + ghdsCalc;
    ghdsEditingCalc = ghdsEditing + ghdsCalc;

    { TFieldType sets }
    ghftAll = [Low (TFieldType)..High (TFieldType)];
    ghftKnown = ghftAll - [ftUnknown];
    ghftInts = [ftSmallint, ftInteger, ftWord, ftAutoInc, ftLargeInt];
    ghftFloats = [ftFloat, ftCurrency, ftBCD, ftFMTBcd];
    ghftNumerics = ghftInts + ghftFloats;

    { Update Status }
    ghusChanges = [usModified, usInserted, usDeleted];

  Type
    { Data Set Class type }
    TDataSetClass = Class Of TDataSet;

    { Before Continue Field Event type }
    TghBeforeContinueFieldEvent = Procedure (Field :TField;
      Var AContinue :Boolean) Of Object;

    { Change Update Status type }
    TghChangeUpdateStatus = usModified..usDeleted;

    { Change Update Status Set type }
    TghChangeUpdateStatusSet = Set Of TghChangeUpdateStatus;

    { Clone Data Set Event type }
    TghCloneDataSetEvent = Procedure (Sender :TObject;
      DataSet :TDataSet; Var Cls :TDataSetClass;
      Var Clone :TDataSet) Of Object;

    { Data Events type }
    TghDataEvents = Set Of TDataEvent;

    { Data Set Filtering type }
    TghDataSetFiltering = Packed Record
      Active :Boolean;
      Expr :String;
      Event :TFilterRecordEvent;
      Bookmark :TBookmark;
    End;

    TghFieldAfterChangeEvent = Procedure (Field :TField;
      Const PrevValue :Variant) Of Object;

    TghFieldBeforeChangeEvent = Procedure (Field :TField;
      Const NewValue :Variant) Of Object;

    { Field Types type }
    TghFieldTypes = Set Of TFieldType;

    TghmsCloseModal = Packed Record
      ID :Cardinal;
      ModalResult :TModalResult;
      Unused :LongInt;
      Result :LongInt;
    End;

    { Table Error type }
    TghTableError = Packed Record
      Code :Integer;
      Table :String;
      Field :String;
    End;

  { Functions and procedures }

  { Abort }
  Procedure ghAbort (Const FocusControl :TWinControl); Overload;

  { Abort }
  Procedure ghAbort (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil); Overload;

  { Abort if No }
  Procedure ghAbortNo (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);

  { Abort if Yes }
  Procedure ghAbortYes (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);

  { Ask }
  Function ghAsk (Const CaptionOrMsg :String; Const Msg :String = '';
    Options :Integer = mb_YesNo) :Integer;

  { Ask Retry }
  Function ghAskRetry (Const CaptionOrMsg :String; Const Msg :String = '')
    :Integer;

  { Ask Retry or Abort }
  Function ghAskRetryAbort (Const CaptionOrMsg :String;
    Const Msg :String = '') :Integer;

  { Check User Level }
  Procedure ghCheckUserLevel (Const Min :Integer;
    Const CaptionOrMsg :String = ''; Const Msg :String = '');

  { Clear Parameters }
  Procedure ghClearParams (Const Params :TParams);

  { Clone Data Set }
  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Cls :TDataSetClass; Const Owner :TComponent = Nil;
    ExcludedProps :String = 'Active'; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;

  { Clone Data Set }
  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Owner :TComponent = Nil; Const ExcludedProps :String = 'Active';
    Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;

  { Container }
  Function ghContainer (Const Control :TControl;
    Const Cls :TWinControlClass) :TWinControl; Overload;

  { Container }
  Function ghContainer (Const Cls :TWinControlClass) :TWinControl;
    Overload;

  { Copy Data Set Props }
  Procedure ghCopyDataSetProps (Const Source, Dest :TDataSet;
    Const Excludeds :String = ''; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '');

  { Copy Fields }
  Procedure ghCopyFields (Const Source, Dest :TDataSet;
    Const ExcludedProps :String = '');

  { DataSnap Attribute }
  Function ghDataSnapAttr (Const UpdateStatus :TUpdateStatus) :Integer;

  { DataSnap Attributes }
  Function ghDataSnapAttrs (Const UpdateStatus :TUpdateStatusSet) :Integer;

  { DataSnap Update Status }
  Function ghDataSnapUpdateStatus (Const Attr :Integer) :TUpdateStatus;

  { DataSnap Update Status Set }
  Function ghDataSnapUpdateStatusSet (Const Attrs :Integer)
    :TUpdateStatusSet;

  { Equal Fields? }
  Function ghEqualFields (Const List1, List2 :TList) :Boolean;

  { Field Count }
  Function ghFieldCount (Const Fields :String) :Integer;

  { Field Type }
  Function ghFieldType (Const Name :String) :TFieldType;

  { Field Type Extended }
  Function ghFieldTypeEx (Const AType :String) :TFieldType;

  { Field Type Integer }
  Function ghFieldTypeInt (Const Name :String) :Integer;

  { Filtering }
  Function ghFiltering (Const DataSet :TDataSet) :TghDataSetFiltering;

  { Inform }
  Procedure ghInform (Const CaptionOrMsg :String; Const Msg :String = '');

  { Input Parameter }
  Function ghInputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;

  { Input Parameter Count }
  Function ghInputParamCount (Const Params :TParams) :Integer;

  { Is? }
  Function ghIs (Const Window :THandle; Const Cls :TWinControlClass)
    :Boolean;

  { Is Button? }
  Function ghIsButton (Const Window :THandle) :Boolean;

  { Master Link }
  Function ghMasterLink (Const ADataSet :TDataSet) :TMasterDataLink;

  { Modal Level }
  Function ghModalLevel :Integer;

  { Output Parameter }
  Function ghOutputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;

  { Output Parameter Count }
  Function ghOutputParamCount (Const Params :TParams) :Integer;

  { Root Control }
  Function ghRootControl (Const Control :TControl) :TControl;

  { Set Cursor }
  Function ghSetCursor (Const Value :TCursor) :TCursor;

  { Show Message }
  Function ghShowMsg (Const CaptionOrMsg :String; Const Msg :String = '';
    Const Options :Integer = 0; Icon :String = '') :Integer; Overload;

  { Show Message }
  Function ghShowMsg (Const Caption, Msg, Icon :String) :Integer; Overload;

  { Show Stop }
  Procedure ghShowStop (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);

  { SQL }
  Function ghSQL (Const DataSet :TDataSet) :String;

  { SQL List }
  Function ghSQLList (Const DataSet :TDataSet) :TStrings;

  { SQL Table }
  Function ghSQLTable (Const DataSet :TDataSet) :String;

  { Table }
  Function ghTable (Const DataSet :TDataSet) :String;

  { Want Enter? }
  Function ghWantEnter (Const Window :THandle) :Boolean;

  { Warn }
  Function ghWarn (Const CaptionOrMsg :String; Const Msg :String = '';
    Const Options :Integer = 0) :Integer;

Implementation

  Uses
    DBCommon, TypInfo, Forms, SysUtils, Math;

  { Functions and procedures }

  Procedure ghAbort (Const FocusControl :TWinControl); Overload;
  Begin
    FocusControl.SetFocus;
    Abort;
  End;

  Procedure ghAbort (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil); Overload;
  Begin
    ghShowStop (CaptionOrMsg, Msg, FocusControl);
    Abort;
  End;

  Procedure ghAbortNo (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);
  Begin
    If ghAsk (CaptionOrMsg, Msg) = IDNo Then
    Begin
      If FocusControl <> Nil Then
        FocusControl.SetFocus;
        
      Abort;
    End;
  End;

  Procedure ghAbortYes (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);
  Begin
    If ghAsk (CaptionOrMsg, Msg) = IDYes Then
    Begin
      If FocusControl <> Nil Then
        FocusControl.SetFocus;

      Abort;
    End;
  End;

  {}//probar, mb_IconWarning Or mb_IconStop (ghAskRetry)
  Function ghAsk (Const CaptionOrMsg :String; Const Msg :String = '';
    Options :Integer = mb_YesNo) :Integer;
  Begin
    If Not ghHasBit (Options, mb_IconQuestion Or mb_IconWarning) Then
      If (ghIndex (CaptionOrMsg, '!') > 0) Or (ghIndex (Msg, '!') > 0) Then
        Options := Options Or mb_IconWarning
      Else
        If Not ghHasBit (Options, mb_IconMask) Then
          Options := Options Or mb_IconQuestion;

    Result := ghShowMsg (CaptionOrMsg, Msg, Options);
  End;

  {}//probar
  Function ghAskRetry (Const CaptionOrMsg :String; Const Msg :String = '')
    :Integer;
  Begin
    Result := ghAsk (CaptionOrMsg, Msg, mb_RetryCancel Or mb_IconStop);
  End;

  {}//probar
  Function ghAskRetryAbort (Const CaptionOrMsg :String;
    Const Msg :String = '') :Integer;
  Begin
    Result := ghAsk (CaptionOrMsg, Msg,
      mb_AbortRetryIgnore Or mb_IconStop);
  End;

  Procedure ghCheckUserLevel (Const Min :Integer;
    Const CaptionOrMsg :String = ''; Const Msg :String = '');
  Begin
    If GHURCurrent.Level >= Min Then
      Exit;

    If (CaptionOrMsg <> '') Or (Msg <> '') Then
      ghAbort (CaptionOrMsg, Msg)
    Else
      ghAbort;
  End;

  Procedure ghClearParams (Const Params :TParams);
  Var
    I :Integer;
  Begin
    For I := 0 To Params.Count - 1 Do
      Params [I].Clear;
  End;

  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Cls :TDataSetClass; Const Owner :TComponent = Nil;
    ExcludedProps :String = 'Active'; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;
  Var
    Params :TObject;
  Begin
    { We don't copy the IndexName property when the IndexFieldNames
      property has a value (otherwise it would be lost). }
    If ghPropStr (DataSet, 'IndexFieldNames') <> '' Then
      ExcludedProps := ghConcat (ExcludedProps, 'IndexName', ';');

    Result := TDataSet (ghClone (DataSet, Cls, Owner, ExcludedProps));

    If IncludeFields Then
      ghCopyFields (DataSet, Result, ExcludedFieldProps);

    Params := ghPropObj (DataSet, TParams);

    If Params <> Nil Then
      TParams (ghPropObj (Result, TParams)).Assign (TParams (Params));
  End;

  Function ghCloneDataSet (Const DataSet :TDataSet;
    Const Owner :TComponent = Nil; Const ExcludedProps :String = 'Active';
    Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '') :TDataSet; Overload;
  Begin
    Result := ghCloneDataSet (DataSet, TDataSetClass (DataSet.ClassType),
      Owner, ExcludedProps, IncludeFields, ExcludedFieldProps);
  End;

  Function ghContainer (Const Control :TControl;
    Const Cls :TWinControlClass) :TWinControl; Overload;
  Begin
    If Control Is Cls Then
      Result := TWinControl (Control)
    Else
    Begin
      Result := Control.Parent;

      While Not ghIsOrNil (Result, Cls) Do
        Result := Result.Parent;
    End;
  End;

  Function ghContainer (Const Cls :TWinControlClass) :TWinControl;
    Overload;
  Begin
    If Screen.ActiveControl <> Nil Then
      Result := ghContainer (Screen.ActiveControl, Cls)
    Else
      Result := Nil;
  End;

  Procedure ghCopyDataSetProps (Const Source, Dest :TDataSet;
    Const Excludeds :String = ''; Const IncludeFields :Boolean = True;
    Const ExcludedFieldProps :String = '');
  Begin
    ghCopyProps (Source, Dest, Excludeds);

    If IncludeFields Then
      ghCopyFields (Source, Dest, ExcludedFieldProps);
  End;

  Procedure ghCopyFields (Const Source, Dest :TDataSet;
    Const ExcludedProps :String = '');
  Var
    I :Integer;
  Begin
    With Source Do
    Begin
      For I := 0 To Fields.Count - 1 Do
        TField (ghClone (Fields [I], Dest, ExcludedProps)).DataSet := Dest;

      For I := 0 To AggFields.Count - 1 Do
        TField (ghClone (AggFields [I], Dest, ExcludedProps)).DataSet :=
          Dest;
    End;
  End;

  Function ghDataSnapAttr (Const UpdateStatus :TUpdateStatus) :Integer;
  Begin
    Case UpdateStatus Of
      usModified   : Result := dsRecModified;
      usInserted   : Result := dsRecNew;
      usDeleted    : Result := dsRecDeleted;
      Else
        Result := 0;
    End;
  End;

  Function ghDataSnapAttrs (Const UpdateStatus :TUpdateStatusSet) :Integer;
  Var
    Status :TUpdateStatus;
  Begin
    Result := 0;

    For Status := Low (TUpdateStatus) To High (TUpdateStatus) Do
      If Status In UpdateStatus Then
        Result := Result Or ghDataSnapAttr (Status);
  End;

  Function ghDataSnapUpdateStatus (Const Attr :Integer) :TUpdateStatus;
  Begin
    Case Attr Of
      dsRecDeleted    : Result := usDeleted;
      dsRecNew        : Result := usInserted;
      dsRecModified   : Result := usModified;
      Else
        Result := usUnmodified;
    End;
  End;

  Function ghDataSnapUpdateStatusSet (Const Attrs :Integer)
    :TUpdateStatusSet;
  Begin
    Result := [];

    If ghHasBit (Attrs, dsRecDeleted) Then
      Include (Result, usDeleted);

    If ghHasBit (Attrs, dsRecNew) Then
      Include (Result, usInserted);

    If ghHasBit (Attrs, dsRecModified) Then
      Include (Result, usModified);
  End;

  Function ghEqualFields (Const List1, List2 :TList) :Boolean;
  Var
    I :Integer;
  Begin
    Result := List1.Count = List2.Count;

    If Result Then
      For I := 0 To List1.Count - 1 Do
        If TField (List1 [I]).Value <> TField (List2 [I]).Value Then
        Begin
          Result := False;
          Exit;
        End;
  End;

  Function ghFieldCount (Const Fields :String) :Integer;
  Begin
    If Fields <> '' Then
      Result := ghCountTo (Fields, ';', Length (Fields) - 1) + 1
    Else
      Result := 0;
  End;

  Function ghFieldType (Const Name :String) :TFieldType;
  Begin
    Result := TFieldType (Max (ghFieldTypeInt (Name), 0 { ftUnknown }));
  End;

  Function ghFieldTypeEx (Const AType :String) :TFieldType;
  Var
    Value :Integer;
  Begin
    Result := ftUnknown;

    If AType = '' Then
      Exit;

    Value := ghFieldTypeInt (AType);

    If Value = -1 Then
      If ghEquals (AType, 'Double') Then
      Begin
        Result := ftFloat;
        Exit;
      End
      Else
      Begin
        Value := ghFieldTypeInt ('ft' + AType);

        If (Value = -1) And (AType [1] In ['T', 't']) Then
          Value := ghFieldTypeInt ('ft' + ghCopy (AType, 2));
      End;

    If Value > -1 Then
      Result := TFieldType (Value);
  End;

  Function ghFieldTypeInt (Const Name :String) :Integer;
  Begin
    Result := GetEnumValue (TypeInfo (TFieldType), Name);
  End;

  Function ghFiltering (Const DataSet :TDataSet) :TghDataSetFiltering;
  Begin
    With DataSet Do
    Begin
      Result.Active := Filtered;
      Result.Expr := Filter;
      Result.Event := OnFilterRecord;
      Result.Bookmark := GetBookmark;
    End;
  End;

  Procedure ghInform (Const CaptionOrMsg :String; Const Msg :String = '');
  Begin
    ghShowMsg (CaptionOrMsg, Msg, mb_IconInformation);
  End;

  {$Warn No_RetVal Off}
  Function ghInputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;
  Var
    I :Integer;
  Begin
    For I := 0 To MaxInt Do
      If Params [I].ParamType In [ptInput, ptInputOutput] Then
      Begin
        Dec (Index);

        If Index = -1 Then
        Begin
          Result := Params [I];
          Exit;
        End;
      End;
  End;
  {$Warn No_RetVal On}

  Function ghInputParamCount (Const Params :TParams) :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Params.Count - 1 Do
      If Params [I].ParamType In [ptInput, ptInputOutput] Then
        Inc (Result);
  End;

  Function ghIs (Const Window :THandle; Const Cls :TWinControlClass)
    :Boolean;
  Begin
    Result := FindControl (Window) Is Cls;
  End;

  Function ghIsButton (Const Window :THandle) :Boolean;
  Begin
    // To determinate when a Windows control is a push button

    Result := ghIs (Window, TButton) Or

      { NOTE: Some third party controls do not respond to the wm_GetDlgCode
        message properly. }
      ghGetDialogCodeButton (Window) Or

      ghClassButton (Window);
  End;

    Type
      TDataSourceAccess = Class (TDataSource);
      TMasterDataLinkAccess = Class (TMasterDataLink);
  Function ghMasterLink (Const ADataSet :TDataSet) :TMasterDataLink;
  Var
    I :Integer;
  Begin
    { In some data set classes, the master data link is a private field
      without access property.  This function returns that data link. }

    If ADataSet.DataSource <> Nil Then
      With TDataSourceAccess (ADataSet.DataSource) Do
        For I := 0 To DataLinks.Count - 1 Do
          If (TDataLink (DataLinks [I]) Is TMasterDataLink) And
          (TMasterDataLinkAccess (DataLinks [I]).GetDetailDataSet =
          ADataSet) Then
          Begin
            // DataSet's master data link found
            Result := DataLinks [I];
            Exit;
          End;

    Result := Nil;
  End;

  Function ghModalLevel :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Screen.CustomFormCount - 1 Do
      If fsModal In Screen.CustomForms [I].FormState Then
        Inc (Result);
  End;

  {$Warn No_RetVal Off}
  Function ghOutputParam (Const Params :TParams; Index :Integer = 0)
    :TParam;
  Var
    I :Integer;
  Begin
    For I := 0 To MaxInt Do
      If Params [I].ParamType In [ptOutput, ptInputOutput, ptResult] Then
      Begin
        Dec (Index);

        If Index = -1 Then
        Begin
          Result := Params [I];
          Exit;
        End;
      End;
  End;
  {$Warn No_RetVal On}

  Function ghOutputParamCount (Const Params :TParams) :Integer;
  Var
    I :Integer;
  Begin
    Result := 0;

    For I := 0 To Params.Count - 1 Do
      If Params [I].ParamType In [ptOutput, ptInputOutput, ptResult] Then
        Inc (Result);
  End;

  Function ghRootControl (Const Control :TControl) :TControl;
  Begin
    Result := Control;

    While (Result.Parent <> Nil) And Not (Result.Parent Is TCustomForm) Do
      Result := Result.Parent;
  End;

  Function ghSetCursor (Const Value :TCursor) :TCursor;
  Begin
    Result := Screen.Cursor;
    Screen.Cursor := Value;
  End;

  Function ghShowMsg (Const CaptionOrMsg :String; Const Msg :String = '';
    Const Options :Integer = 0; Icon :String = '') :Integer; Overload;
  Var
    Params :TMsgBoxParams;
    Sound :Integer Absolute Result;
    WindowList :Pointer;
  Begin
    Params.cbSize := ghZero (@Params, SizeOf (Params));
    Params.hwndOwner := Application.Handle;

    If Msg <> '' Then
    Begin
      Params.lpszText := Pointer (Msg);
      Params.lpszCaption := PChar (CaptionOrMsg);
    End
    Else
    Begin
      Params.lpszText := PChar (CaptionOrMsg);
      Params.lpszCaption := PChar (Application.Title);
    End;

    Sound := Options And mb_IconMask;

    If Icon = '' Then
      Case Sound Of
        mb_IconStop : Icon := 'GHMB_ICONSTOP';
        mb_IconQuestion : Icon := 'GHMB_ICONQUESTION';
        mb_IconWarning : Icon := 'GHMB_ICONWARNING';
        mb_IconInformation : Icon := 'GHMB_ICONINFORMATION';
      End;

    If (Icon <> '') And (ghFindResource (Icon, rt_Group_Icon) <> 0) Then
    Begin
      Params.hInstance := HInstance;
      Params.dwStyle := ghTurnOff (Options, mb_IconMask) Or mb_UserIcon;

      If Win32Platform = ver_Platform_Win32_NT Then
        Params.lpszIcon := Pointer (WideString (Icon))
      Else
        Params.lpszIcon := Pointer (Icon);

      If Sound > 0 Then
        MessageBeep (Sound);
    End
    Else
      Params.dwStyle := Options;

    WindowList := DisableTaskWindows (0);

    Try
      Result := Integer (MessageBoxIndirect (Params));
    Finally
      EnableTaskWindows (WindowList);
    End;
  End;

  Function ghShowMsg (Const Caption, Msg, Icon :String) :Integer; Overload;
  Begin
    Result := ghShowMsg (Caption, Msg, 0, Icon);
  End;

  Procedure ghShowStop (Const CaptionOrMsg :String; Const Msg :String = '';
    Const FocusControl :TWinControl = Nil);
  Begin
    If FocusControl <> Nil Then
      FocusControl.SetFocus;

    ghShowMsg (CaptionOrMsg, Msg, mb_IconStop);
  End;

  Function ghSQL (Const DataSet :TDataSet) :String;
  Var
    SQL :TStrings;
  Begin
    SQL := ghSQLList (DataSet);

    If SQL <> Nil Then
      Result := SQL.Text
    Else
      Result := '';
  End;

  Function ghSQLList (Const DataSet :TDataSet) :TStrings;
  Begin
    Result := TStrings (ghPropObj (DataSet, 'SQL', TStrings));
  End;

  Function ghSQLTable (Const DataSet :TDataSet) :String;
  Begin
    Result := ghSQL (DataSet);

    If Result <> '' Then
      Result := GetTableNameFromSQL (Result)
  End;

  Function ghTable (Const DataSet :TDataSet) :String;
  Begin
    If IsPublishedProp (DataSet, 'TableName') Then
      Result := GetPropValue (DataSet, 'TableName', False)
    Else
      Result := ghSQLTable (DataSet)
  End;

  Function ghWantEnter (Const Window :THandle) :Boolean;
  Begin
    Result := ghWantAllKeys (Window) Or ghIsButton (Window);
  End;

  Function ghWarn (Const CaptionOrMsg :String; Const Msg :String = '';
    Const Options :Integer = 0) :Integer;
  Begin
    Result := ghShowMsg (CaptionOrMsg, Msg, Options Or mb_IconWarning);
  End;

End.

